/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef EDGE_H
#define EDGE_H


#include <algorithm>
#include <iostream>
#include "../../runtime/timing.h"

#ifdef NULL
#undef NULL
#endif


// Defines an graph edge for the purpose of verifying time constraint
// viability. An edge contains a minimum and maximum rttime_t.
//   Note that arithmetic operators are defined in order that the edges
// produce correct results when used as elements in matrix multiplication. As
// such, they do not necessarily correspond to the arithmetic operations
// performed on the member data.
struct edge
{
    // Constraints
    rttime_t min;
    rttime_t max;
    // Flags
    bool minset;
    bool maxset;

    // State constants: A NULL edge denotes a relation where no edge exists,
    // UNKNOWN means that the constraints cannot be verified
    const static int NULL=0, UNKNOWN=1, NORMAL=2;
    int state;

    edge()
    : min(rt_inf), max(rt_zero), minset(false), maxset(false), state(NULL)
    { }

    // Create a normal edge
    explicit edge(int dummy)
//    : min(rt_zero), max(rt_inf), minset(false), maxset(false), state(NORMAL) { }
    : min(rt_zero), max(rt_inf), minset(true), maxset(true), 
state(NORMAL) {}

    // Create a normal edge from the specified time constraints
    edge(rttime_t mn, rttime_t mx)
//    : min(mn), max(mx), minset(false), maxset(false), state(NORMAL) { }
    : min(mn), max(mx), minset(true), maxset(true), state(NORMAL) { }

    static edge unknown()
    {
        edge e;
        e.state = UNKNOWN;
        return e;
    }

    // Overridden operators to facilitate necessary math operations when
    // multiplying matrices.

    // + represents selection: Select the proper edge constraints
    // (a constraint that may choose two different paths)
    const edge& operator+=(const edge& e);

    // * represents addition (a constraint over several consecutive edges)
    const edge& operator*=(const edge& e);
};


inline edge operator+(const edge& e1, const edge& e2)
{
    edge res(e1);
    res += e2;
    return res;
}

inline edge operator*(const edge& e1, const edge& e2)
{
    edge res(e1);
    res *= e2;
    return res;
}

inline bool operator==(const edge& e1, const edge& e2)
{
    return e1.min == e2.min and e1.max == e2.max;
}


std::ostream& operator<<(std::ostream& os, const edge& e);


#endif // #define EDGE_H
