/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <algorithm>
#include <fstream>
#include <iostream>
#include <map>
#include <sstream>
#include <string>
#include <vector>

#include <boost/tuple/tuple.hpp>

#include "edge.h"
#include "matrix.h"

// (actor, method)
typedef std::pair<std::string, std::string> method_t;
// (type, expr, synchronizer)
typedef boost::tuple<std::string, std::string, std::string> cost_tuple;


std::ostream& operator<<(std::ostream& os, const method_t m)
{
    os << m.first << "." << m.second;
    return os;
}

// A simple function to tell if a string is numeric or not. I'm fairly
// certain a function like this already exists, but I do not know it.
bool is_numeric(std::string s){
    for(int i = 0; i < (int)s.size(); i++){
        char c = s[i];
        if(c < '0' || c > '9')
            return false;
    }
    return true;
}

// Parse expressions; should at some point be able to parse
// constant arithmetic expressions but discard variables as unknown
rttime_t doparse(const std::string& s, const rttime_t& def)
{
    int i;
    std::stringstream ss(s);

	std::cout <<s <<std::endl;

    if((ss >> i))
    {
        return rt_zero + rttime_t(0, i);
    }
    else
        return def;
}

// Parse constraint expressions to determine an edge
// Expects a list of all (0, 1, or 2) constraints for the
// edge in question
edge parse_expr(const std::vector<cost_tuple>& tlist)
{
    using namespace std;
    using namespace boost;

    if(tlist.size() == 0)
        return edge::unknown();

    edge e(0);
    for(vector<cost_tuple>::const_iterator it = tlist.begin();
        it != tlist.end(); ++it)
    {
        if(get<0>(*it) == "min")
        {
            e.min = doparse(get<1>(*it), rt_zero);
            e.minset = true;
        }
        else // max constraint
        {
            e.max = doparse(get<1>(*it), rt_inf);
            e.maxset = true;
        }
    }
    return e;
}

// Performs a static 'sanity check' on the generated code.  Checks to ensure
// that easily-visible, impossible constraints have not been entered.
// For example, a must-run-before (max constraint) that is sooner than a must-
// run-after (min).
//   Also, checks to ensure that no more than 1 max and 1 min constraint can
// be entered on any one relationship.
//   Returns true if the input is good, false if it is not, in  which case
// error messages are printed to stderr.
bool sanitycheck(std::istream& in)
{
    using namespace std;
    using namespace boost;

    // Matrix index by method
    map<method_t, int> index_map;
    // Method by index
    vector<method_t> method_list;
    // Calls made from each method (node)
    map<method_t, vector<method_t> > call_map;
    // Cost for each edge
    typedef map<pair<method_t, method_t>, vector<cost_tuple> > costmap_t;
    costmap_t cost_map, global_map;

    string s; // buffer
    int index = 0; // last matrix index used

    // Parse the graph.src file generated by the compiler to form a graph.
    // NOTE: graph.src use should be discontinued after debugging and replaced
    // with a stringstream.
    while((in >> s))
    {
        if(s == "<method>")
        {
            string actor, method;
            in >> actor >> method;
            method_t cur_method(actor, method);

            // Which row does this method correspond to?
            index_map[cur_method] = index;
            // Which method does this index refer to?
            if(index >= (int) method_list.size())
                method_list.resize(index+1);
            method_list[index] = cur_method;
            index++; // moving on

            call_map[cur_method] = vector<method_t>(); // make sure it's there

            // Push back calls until the method ends
            while((in >> s) && s != "</method>")
            {
                int numconditions;
                stringstream ss(s);
                ss >> numconditions;

                for(int i = 0; i < numconditions; ++i)
                {
                    string type, expr;
                    in >> type;
                    getline(in, expr);

                }

                string c_actor;
                if(is_numeric(s)){
                    in >>c_actor;
                }else{
                    c_actor = s;
                }

                //string c_actor;
                //in >> c_actor;
                string c_method;
                in >> c_method;
                call_map[cur_method].push_back(method_t(c_actor, c_method));
            }
        }
        else if(s == "<sync>")
        {
            string sync;
            in >> sync;

            while((in >> s) && s != "</sync>")
            {
                string& type = s;
                string s_actor, s_method, d_actor, d_method;
                in >> s_actor >> s_method >> d_actor >> d_method;
                string expr;

                in.ignore(100, '\n');
                getline(in, expr);

                method_t src(s_actor, s_method), dest(d_actor, d_method);

                //cout <<"added call from " <<src <<" to " <<dest <<endl;

                cost_map[pair<method_t, method_t>(src, dest)].push_back(
                    make_tuple(type, expr, sync));
            }
        }
        else if(s == "<globalsync>")
        {
            while((in >> s) && s != "</globalsync>")
            {
                string& type = s;
                string s_actor, s_method, d_actor, d_method;
                in >> s_actor >> s_method >> d_actor >> d_method;
                string expr;

                in.ignore(100, '\n');
                getline(in, expr);

                method_t src(s_actor, s_method), dest(d_actor, d_method);

                global_map[pair<method_t, method_t>(src, dest)].push_back(
                    make_tuple(type, expr, "global"));
            }
        }
    } // done reading source file

    bool edges_consistent = true;

    // Set up the matrix
    cout <<call_map.size() <<endl;
    matrix<edge> ematrix(call_map.size(), call_map.size());
    map<method_t, vector<method_t> >::iterator map_it = call_map.begin();
    // For each node (method)
    while(map_it != call_map.end())
    {
        const method_t& lhs = map_it->first; // caller

        cout <<"checking calls from " <<lhs.first <<"." <<lhs.second
             <<"()" <<endl;

        for(vector<method_t>::iterator it = map_it->second.begin();
            it != map_it->second.end(); ++it)
        {
            cout <<"\tcall to " <<it->first <<"." <<it->second <<"()"
                 <<endl;
            // Same actor? Internal method calls are not constrained
         
            if(lhs.first == it->first)
            {
                ematrix(index_map[lhs], index_map[*it]) = edge(0);
            }
            else // figure out the cost 
            {
                vector<cost_tuple> v = cost_map[pair<method_t, method_t>(lhs,
                    *it)];

                cout <<"\twith " <<v.size() <<" constraints:" <<endl;

                // An edge can -at most- be determined by two constraints,
                // where one must be min and one max; more or identical is
                // an error
                if(v.size() > 2 ||
                   (v.size() == 2 && get<0>(v[0]) == get<0>(v[1])))
                {
                    cout << "Error in synchronizer " << get<2>(v[0]) << ":\n"
                         << "Too many constraints defined between "
                         << lhs << " and " << *it << endl;
                    return false;
                }

                edge e = parse_expr(v);
                if(e.minset && e.maxset)
                {
                    if(e.min > e.max)
                    {
                        cout << "Error: Constraints between "
                            << lhs << " and " << *it
                            << " are inconsistent\n(run before < run after)"
                            << endl;

                        edges_consistent = false;
                    }
                    if(e.min == e.max)
                    {
                        cout << "Warning: Constraints between "
                            << lhs << " and " << *it
                            << " cannot practically be met\n"
                            << "(run before = run after)"
                            << endl;
                    }
                }
cout <<e <<endl;
                if(edges_consistent)
                    ematrix(index_map[lhs], index_map[*it]) = e;

                edges_consistent = true;
            }
        }
        ++map_it;
    } // done initialising matrix

    if(!edges_consistent)
        return false;

    // Set diagonal to imposible times (min = inf, max = 0)
    for(int i = 0; i < ematrix.ROWS; ++i)
        ematrix(i, i) = edge(rt_inf, rt_zero);


    //cout << ematrix << endl;

    const matrix<edge> matrix_inc(ematrix);
    matrix<edge> endmatrix(ematrix);

    // Set up the result matrix
    for(unsigned row = 0; row < call_map.size(); ++row)
        for(unsigned col = 0; col < call_map.size(); ++col)
            if(endmatrix(row, col).state != edge::NORMAL)
                endmatrix(row, col) = edge(rt_inf, rt_zero);

    // Perform the math to obtain the transitive, reflexive closure of the graph
    for(int i = 0; i < ematrix.ROWS; ++i)
    {
        ematrix *= matrix_inc;

        for(unsigned row = 0; row < call_map.size(); ++row)
        {
            for(unsigned col = 0; col < call_map.size(); ++col)
            {
                // only copy value to result matrix if new min < result min
                if(ematrix(row, col).min < endmatrix(row, col).min)
                    endmatrix(row, col).min = ematrix(row, col).min;
                // only copy value to result matrix if new max > result max
                if(ematrix(row, col).max > endmatrix(row, col).max)
                    endmatrix(row, col).max = ematrix(row, col).max;
            }
        }

        //cout << ematrix << endl;
    }

    // PROBLEM: Indirect max constraints may grow unboundedly due to loops. This
    // can happen, and arbitrarily large global constraints should then be
    // allowed! We must therefore find out what paths -are- cyclic and set the
    // corresponding max constraints to infinity. Min constraints are not an
    // issue since optimal min paths are by necessity non-cyclic.
    //   To find out what paths are cyclic we simply square the matrix and see
    // what paths would increase their max constraints; these are more than n
    // steps and must contain cycles.
    ematrix *= ematrix;
    for(unsigned row = 0; row < call_map.size(); ++row)
    {
        for(unsigned col = 0; col < call_map.size(); ++col)
        {
            if(ematrix(row, col).max > endmatrix(row, col).max)
                endmatrix(row, col).max = rt_inf;
        }
    }

    cout << "\nReflexive transitive closure of edge matrix..." << endl;
    cout << endmatrix << endl;

    // Perform static check on global constraints
    for(costmap_t::iterator it = global_map.begin(); it != global_map.end(); ++it)
    {
        const method_t& src = it->first.first, dest = it->first.second;
        const vector<cost_tuple> cost = it->second;

        edge e = parse_expr(cost);
        if(e.minset && e.maxset)
        {
            if(e.min > e.max)
            {
                cout << "Error: Global constraints between "
                    << src << " and " << dest
                    << " are inconsistent\n(run before < run after)"
                    << endl;
                return false;
            }
            if(e.min == e.max)
            {
                cout << "Warning: Global constraints between "
                    << src << " and " << dest
                    << " cannot practically be met\n"
                    << "(run before = run after)"
                    << endl;
            }
        }

        const edge& indirect = endmatrix(index_map[src], index_map[dest]);

        if(e.minset)
        {
            if(e.min > indirect.max)
            {
                cout <<e.min <<" > " <<indirect.max <<endl;
                cout << "Error: Global constraints between "
                    << src << " and " << dest
                    << " cannot be met\n(run before < run after)"
                    << endl;

                return false;
            }
            else if(e.min == indirect.max)
            {
                cout << "Warning: Global constraints between "
                    << src << " and " << dest
                    << " cannot practically be met\n"
                    << "(run before = run after)"
                    << endl;
            }
        }
        if(e.maxset)
        {
            if(e.max < indirect.min)
            {
                cout << "Error: Global constraints between "
                    << src << " and " << dest
                    << " cannot be met\n(run before < run after)"
                    << endl;

                return false;
            }
            else if(e.max == indirect.min)
            {
                cout << "Warning: Global constraints between "
                    << src << " and " << dest
                    << " cannot practically be met\n"
                    << "(run before = run after)"
                    << endl;
            }
        }
    }

    return true;
}
