/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef MATRIX_H
#define MATRIX_H


#include <cassert>
#include <iostream>
#include <vector>


// Templated matrix class, written because no suitable third-party matrix
// library was found (that is, we would like fast matrix multiplication and
// require the ability to insert arbitrary data types as elements).
//   Elements are access via operator(row, col).
template <typename elem> class matrix
{
    std::vector<std::vector<elem> > data;
public:
    const int ROWS, COLS;

    matrix(int rows, int cols) : ROWS(rows), COLS(cols)
    {
        data.resize(ROWS);
        for(int i = 0; i < ROWS; ++i)
            data[i].resize(COLS);
    }

    matrix(const matrix<elem> &m)
    : data(m.data), ROWS(m.ROWS), COLS(m.COLS)
    { }

    // Assignment operator. Matrices must have identical dimensions!
    const matrix<elem>& operator=(const matrix<elem> &m)
    {
        using namespace std;
        assert(ROWS == m.ROWS and COLS == m.COLS);
        data = m.data;
        return *this;
    }

    matrix<elem> operator*(const matrix<elem>& m) const
    {
        if(COLS != m.ROWS)
            throw "Matrix multiplication error: Wrong dimensions";

        matrix<elem> res(ROWS, m.COLS);

        for(int i = 0; i < ROWS; ++i)
        {
            for(int j = 0; j < m.COLS; ++j)
            {
                elem sum = data[i][0] * m.data[0][j];
                for(int k = 1; k < COLS; ++k)
                    sum += data[i][k] * m.data[k][j];
                res.data[i][j] = sum;
            }
        }
        return res;
    }

    const matrix<elem>& operator*=(const matrix<elem>& m)
    {
        *this = *this * m;
        return *this;
    }

    elem& operator()(int row, int col)
    {
        return data[row][col];
    }

    const elem& operator()(int row, int col) const
    {
        return data[row][col];
    }
};

template <typename T>
std::ostream& operator<<(std::ostream& os, const matrix<T>& m)
{
    for(int i = 0; i < m.ROWS; ++i)
    {
        os << "[ ";
        for(int j = 0; j < m.COLS; ++j)
            os << m(i,j) << "\t";
        os << "]\n";
    }
    return os;
}

#endif // #define MATRIX_H
