/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

const char* rtsync_version = "1.0-rc3";

#include "error/graphmaker.h"
#include "visitors/ast.h"
#include "visitors/codevisitor.h"
#include "visitors/scopevisitor.h"

#include <cstdio>
#include <iostream>
#include <fstream>
#include <unistd.h>

using namespace std;

// External variables from the YACC parser
extern FILE* yyin;
extern "C" int yyparse(void);

// Root pointer for the AST
node* root = 0;

extern int yyparse();

int main(int argc, char** argv)
{
    using namespace std;

    if(argc < 2){
	cout <<"rtsync: no input files" <<endl;
        exit(1);
    }

    string outfile = "rt_output.cc";

    // Parse command line parameters
    extern char* optarg;
    extern int optind;
    int opt;
    while((opt = getopt(argc, argv, "vo:")) != -1)
    {
        switch(opt)
            {
            case 'o': // output file
                {
                    outfile = string(optarg);
                    break;
                }
            case 'v': // version info
                {
                    cout << "RTSync version " << rtsync_version << ".\n"
                         << "Copyright (C) 2004 Stefan D. Bruda, Petter Haggholm, and Scott Stoddard.\n"
                         << "This is free software; see the source for copying conditions.  There is NO\n"
                         << "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n";
                    return 0;
                }
            default:
                break;
            }
    }

    char* input = argv[optind];

    // Invoke YACC parser
    yyin = fopen(input, "r");
    int res = yyparse();
    fclose(yyin);

    if(res)
    {
        cerr << "Parser failed with status " << res << endl;
        return 1;
    }

    // Output source to 'outfile' and the graph source to graph.src
    ofstream src(outfile.c_str());
    ofstream of("graph.src");

    // Let a code generating visitor traverse the tree
    codevisitor cv(src, of);
    root->accept(0, cv);

    // Let a scope checking visitor check the tree
    scopevisitor sv;
    root->accept(0, sv);
    if(!sv.good())
    {
        cout << "Error, bailing out ..." << endl;
        return 1;
    }
    of.close();

    // Perform a sanity check on the calling graph
    ifstream in("graph.src");
    if(!sanitycheck(in))
        return 1;

    delete root;
}
