/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "ast.h"

// Ugly, messy list of constructors and destructors. Don't look at this ...
// There should be very little reason to modify this file: All it contains are
// constructors that initialize nodes with the data passed into the
// constructor, and destructors that recursively delete the AST.

program_node::program_node(int line, module_list_node* m)
: node(line), modules(m) { }
program_node::~program_node() { delete modules; }

type_spec_node::type_spec_node(int line, keyword_t t) : node(line), type(t) { }


module_list_node::module_list_node(int line, module_list_node* h, module_node* t)
: node(line), head(h), tail(t) { }
module_list_node::~module_list_node() { delete head; delete tail; }

actor_node::actor_node(int line, char* s, decl_list_node* t)
: module_node(line), id(s), tail(t) { }
actor_node::~actor_node() { free(id); delete tail; }

sync_decl_list_node::sync_decl_list_node(int line,
      var_decls_node* v, sync_init_node* i,
      constraint_block_node* c, trigger_block_node* t, exception_block_node* e)
      : node(line), vars(v), init(i), constraints(c), triggers(t), exception(e) { }
sync_decl_list_node::~sync_decl_list_node()
{ delete vars; delete init; delete constraints; delete triggers; delete exception; }

var_decls_node::var_decls_node(int line, var_decls_node* h, var_decl_node* t)
: node(line), head(h), tail(t) { }
var_decls_node::~var_decls_node() { delete head; delete tail; }

exception_block_node::exception_block_node(int line, compound_stmt_node* b)
: node(line), body(b) { }
exception_block_node::~exception_block_node() { delete body; }

compound_stmt_node::compound_stmt_node(int line, loc_decl_node* d, stmt_list_node* b)
: stmt_node(line), decls(d), body(b) { }
compound_stmt_node::~compound_stmt_node() { delete decls; delete body; }

init_node::init_node(int line, compound_stmt_node* h) : node(line), head(h) { }
init_node::~init_node() { delete head; }

decl_list_node::decl_list_node(int line, var_decls_node* v, init_node* i,
func_decls_node* m) : node(line), vars(v), init(i), methods(m) { }
decl_list_node::~decl_list_node() { delete vars; delete init; delete methods; }

simple_var_decl_node::simple_var_decl_node(int line, type_spec_node* t, char* s)
: var_decl_node(line), type(t), id(s) { }
simple_var_decl_node::~simple_var_decl_node() { delete type; free(id); }

arr_var_decl_node::arr_var_decl_node(int line, type_spec_node* t, char* s, char* e)
: var_decl_node(line), type(t), id(s), elems(e) { }
arr_var_decl_node::~arr_var_decl_node() { delete type; free(id); free(elems); }

fun_decl_node::fun_decl_node(int line, char* s, params_node* p, compound_stmt_node* b)
: decl_node(line), id(s), params(p), body(b) { }
fun_decl_node::~fun_decl_node() { free(id); delete params; delete body; }

param_list_node::param_list_node(int line, param_list_node* h, param_node* t)
: params_node(line), head(h), tail(t) { }
param_list_node::~param_list_node() { delete head; delete tail; }

simple_param_node::simple_param_node(int line, type_spec_node* t, char* s)
: param_node(line), type(t), id(s) { }
simple_param_node::~simple_param_node() { delete type; free(id); }

arr_param_node::arr_param_node(int line, type_spec_node* t, char* s)
: param_node(line), type(t), id(s) { }
arr_param_node::~arr_param_node() { delete type; free(id); }

loc_decl_node::loc_decl_node(int line, loc_decl_node* h, var_decl_node* t) : node(line), head(h), tail(t) { }
loc_decl_node::~loc_decl_node() { delete head; delete tail; }

stmt_list_node::stmt_list_node(int line, stmt_list_node* h, stmt_node* t) : node(line), head(h), tail(t) { }
stmt_list_node::~stmt_list_node() { delete head; delete tail; }

expr_stmt_node::expr_stmt_node(int line, expr_node* e)
: stmt_node(line), expr(e) { }
expr_stmt_node::~expr_stmt_node() { delete expr; }

select_stmt_node::select_stmt_node(int line, expr_node* e, stmt_node* ifs, stmt_node* els)
: stmt_node(line), expr(e), ifstmt(ifs), elsestmt(els) { }
select_stmt_node::~select_stmt_node() { delete expr; delete ifstmt; delete elsestmt; }

iter_stmt_node::iter_stmt_node(int line, expr_node* e, stmt_node* b)
: stmt_node(line), expr(e), body(b) { }
iter_stmt_node::~iter_stmt_node() { delete expr; delete body; }

return_stmt_node::return_stmt_node(int line, expr_node* e)
: stmt_node(line), expr(e) { }
return_stmt_node::~return_stmt_node() { delete expr; }

call_stmt_node::call_stmt_node(int line, char* s1, char* s2, args_node* a)
: stmt_node(line), id1(s1), id2(s2), args(a) { }
call_stmt_node::~call_stmt_node() { free(id1); free(id2); delete args; }

sync_call_stmt_node::sync_call_stmt_node(int line, char* s1, char* s2, args_node* a)
: stmt_node(line), id1(s1), id2(s2), args(a) { }
sync_call_stmt_node::~sync_call_stmt_node() { free(id1); free(id2); delete args; }

assign_expr_node::assign_expr_node(int line, var_node* v, expr_node* e)
: expr_node(line), var(v), expr(e) { }
assign_expr_node::~assign_expr_node() { delete var; delete expr; }

simple_var_node::simple_var_node(int line, char* s)
: var_node(line), id(s) { }


arr_var_node::arr_var_node(int line, char* s, expr_node* e)
: var_node(line), id(s), expr(e) { }
arr_var_node::~arr_var_node() { free(id); delete expr; }

compound_simple_expr_node::compound_simple_expr_node(int line,
    add_expr_node* e1, relop_node* r, add_expr_node* e2)
: simple_expr_node(line), expr1(e1), rel(r), expr2(e2) { }
compound_simple_expr_node::~compound_simple_expr_node()
{ delete expr1; delete rel; delete expr2; }

relop_node::relop_node(int line, relop_t o) : node(line), op(o) { }


compound_add_expr_node::compound_add_expr_node(int line,
    add_expr_node* e1, addop_node* o, term_node* e2)
: add_expr_node(line), expr1(e1), op(o), expr2(e2) { }
compound_add_expr_node::~compound_add_expr_node() { delete expr1; delete op; delete expr2; }

addop_node::addop_node(int line, addop_t o) : node(line), op(o) { }


compound_term_node::compound_term_node(int line, term_node* e1, mulop_node* o, factor_node* e2)
: term_node(line), expr1(e1), op(o), expr2(e2) { }
compound_term_node::~compound_term_node() { delete expr1; delete op; delete expr2; }

mulop_node::mulop_node(int line, mulop_t o)
: node(line), op(o) { }


paren_factor_node::paren_factor_node(int line, expr_node* e)
: factor_node(line), expr(e) { }
paren_factor_node::~paren_factor_node() { delete expr; }

var_factor_node::var_factor_node(int line, var_node* v)
: factor_node(line), var(v) { }
var_factor_node::~var_factor_node() { delete var; }

num_factor_node::num_factor_node(int line, char* v, int t)
: factor_node(line,t), val(v) { }

call_node::call_node(int line, char* s1, args_node* a)
: factor_node(line), id1(s1), args(a) { }
call_node::~call_node() { free(id1); delete args; }

args_node::args_node(int line, arg_list_node* a) : node(line), args(a) { }
args_node::~args_node() { delete args; }

arg_list_node::arg_list_node(int line, arg_list_node* h, expr_node* t)
: node(line), head(h), tail(t) { }
arg_list_node::~arg_list_node() { delete head; delete tail; }

sync_node::sync_node(int line, char* s, sync_decl_list_node* t)
: module_node(line), id(s), tail(t) { }
sync_node::~sync_node() { free(id); delete tail; }

global_sync_node::global_sync_node(int line, sync_decl_list_node* t)
: module_node(line), tail(t) { }
global_sync_node::~global_sync_node() { delete tail; }

sync_init_node::sync_init_node(int line, actor_params_node* h, compound_stmt_node* t)
: node(line), head(h), tail(t) { }
sync_init_node::~sync_init_node() { delete head; delete tail; }

constraint_block_node::constraint_block_node(int line, constraint_list_node* h)
: node(line), head(h) { }
constraint_block_node::~constraint_block_node() { delete head; }

constraint_list_node::constraint_list_node(int line, constraint_list_node* h,
     constraint_node* t) : node(line), head(h), tail(t) { }
constraint_list_node::~constraint_list_node() { delete head; delete tail; }

constraint_node::constraint_node(int line, int ct, char* a1, char* m1,
                                 char* a2, char* m2, expr_node* t)
: node(line), type(ct), actor1(a1), method1(m1), actor2(a2), method2(m2), tail(t)
{ }
constraint_node::~constraint_node() {
    free(actor1);
    free(method1);
    free(actor2);
    free(method2);
    delete tail;
}

trigger_block_node::trigger_block_node(int line, trigger_list_node* h) : node(line), head(h) { }
trigger_block_node::~trigger_block_node() { delete head; }

trigger_list_node::trigger_list_node(int line, trigger_list_node* h,
     trigger_node* t) : node(line), head(h), tail(t) { }
trigger_list_node::~trigger_list_node() { delete head; delete tail; }

ed_trigger_node::ed_trigger_node(int line, int t, char* a, char* m, expr_node* e)
: trigger_node(line), type(t), actor(a), method(m), expr(e) { }
ed_trigger_node::~ed_trigger_node() { free(actor); free(method); delete expr; }

action_trigger_node::action_trigger_node(int line, char* a, char* m, compound_stmt_node* c)
: trigger_node(line), actor(a), method(m), tail(c) { }
action_trigger_node::~action_trigger_node() { free(actor); free(method); delete tail; }

actor_params_node::actor_params_node(int line, actor_params_node* h, actor_param_node* t)
: node(line), head(h), tail(t) { }
actor_params_node::~actor_params_node() { delete head; delete tail; }

actor_param_node::actor_param_node(int line, char* i1, char* i2)
: node(line), id1(i1), id2(i2) { }


func_decls_node::func_decls_node(int line, func_decls_node* h, fun_decl_node* t)
: node(line), head(h), tail(t) { }
func_decls_node::~func_decls_node() { delete head; delete tail; }

pass_node::pass_node(int line, char* txt)
: module_node(line), text(txt) { }
