/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef AST_H
#define AST_H

#include "ast_decls.h"
#include "visitor.h"
#include <string>


// Defines classes for the AST. This is an ugly file (some parts had to be
// reordered due to unexpected dependencies); for a much cleaner and more
// readable representation of the type hierarchy, refer to the YACC
// specification of the language grammar (this is after all merely a
// collection of types for its implementation).
//   The only interesting note here is that the nodes are all used to
// implement the Visitor Pattern, hence they all accept() visitors.


struct node
{
    int line;

    node(int l) : line(l) { }
    virtual ~node() { }
    // Accept a visitor; use specified nesting/indentation depth for
    // code generation
    virtual void accept(int depth, visitor& v) = 0;
};

struct program_node : node
{
    module_list_node* modules;
    
    program_node(int line, module_list_node* m);
    ~program_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct module_node : node
{
    module_node(int line) : node(line) { }
};

struct actor_node : module_node
{
    char* id;
    decl_list_node* tail;

    actor_node(int line, char* s, decl_list_node* t);
    ~actor_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct sync_decl_list_node : node
{
    var_decls_node* vars;
    sync_init_node* init;
    constraint_block_node* constraints;
    trigger_block_node* triggers;
    exception_block_node* exception;
    
    sync_decl_list_node(int line, var_decls_node* v, sync_init_node* i, 
                        constraint_block_node* c, trigger_block_node* t,
                        exception_block_node* e);
    ~sync_decl_list_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};



struct stmt_node : node
{
    stmt_node(int line) : node(line) { }
};

struct decl_node : node
{
    decl_node(int line) : node(line) { }
};

struct expr_node : node
{
    static const int EX_INV = 0;
    static const int EX_INT = 1;
    static const int EX_FLOAT = 2;
    static const int EX_STRING = 3;
    static const int EX_UNKNOWN = 666;
    static const int EX_ARRAY = 0x8000; // for bitmasking
    
    int type;
    
    expr_node(int line, int t = EX_UNKNOWN) : node(line), type(t) { }
};

struct simple_expr_node : expr_node
{
    simple_expr_node(int line, int t = expr_node::EX_UNKNOWN) : expr_node(line, t) { }
};

struct add_expr_node : simple_expr_node
{
    add_expr_node(int line, int t = expr_node::EX_UNKNOWN) : simple_expr_node(line, t) { }
};

struct term_node : add_expr_node
{
    term_node(int line, int t = expr_node::EX_UNKNOWN) : add_expr_node(line,t) { }
};

struct factor_node : term_node
{
    char* text;
    
    factor_node(int line, int t = expr_node::EX_UNKNOWN) : term_node(line, t) { }
    ~factor_node() { free(text); }
};

struct var_decl_node : decl_node
{
    var_decl_node(int line) : decl_node(line) { }
};

struct params_node : node
{
    params_node(int line) : node(line) { }
};

struct param_node : node
{
    param_node(int line) : node(line) { }
};

struct trigger_node : node
{
    trigger_node(int line) : node(line) { }
};

struct var_node : factor_node
{
    int type;
    var_node(int line) : factor_node(line) { }
};

enum keyword_t { TK_VOID, TK_INT, TK_FLOAT, TK_STRING };
struct type_spec_node : node
{
    keyword_t type;

    type_spec_node(int line, keyword_t t);
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct module_list_node : node
{
    module_list_node* head;
    module_node* tail;
    
    module_list_node(int line, module_list_node* h, module_node* t);
    ~module_list_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct var_decls_node : node
{
    var_decls_node* head;
    var_decl_node* tail;
    
    var_decls_node(int line, var_decls_node* h, var_decl_node* t);
    ~var_decls_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct call_stmt_node : stmt_node
{
    char* id1;
    char* id2;
    args_node* args;
    
    call_stmt_node(int line, char* s1, char* s2, args_node* a);
    ~call_stmt_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct sync_call_stmt_node : stmt_node
{
    char* id1;
    char* id2;
    args_node* args;

    sync_call_stmt_node(int line, char* s1, char* s2, args_node* a);
    ~sync_call_stmt_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct compound_stmt_node : stmt_node
{
    loc_decl_node* decls;
    stmt_list_node* body;

    compound_stmt_node(int line, loc_decl_node* d, stmt_list_node* b);
    ~compound_stmt_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct init_node : node
{
    compound_stmt_node* head;

    init_node(int line, compound_stmt_node* h);
    ~init_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct decl_list_node : node
{
    var_decls_node* vars;
    init_node* init;
    func_decls_node* methods;
    
    decl_list_node(int line, var_decls_node* v, init_node* i, func_decls_node* m);
    ~decl_list_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct simple_var_decl_node : var_decl_node
{
    type_spec_node* type;
    char* id;
    
    simple_var_decl_node(int line, type_spec_node* t, char* s);
    ~simple_var_decl_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct arr_var_decl_node : var_decl_node
{
    type_spec_node* type;
    char* id;
    char* elems;
    
    arr_var_decl_node(int line, type_spec_node* t, char* s, char* e);
    ~arr_var_decl_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct exception_block_node : node
{
    compound_stmt_node* body;
    
    exception_block_node(int line, compound_stmt_node* b);
    ~exception_block_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct fun_decl_node : decl_node
{
    char* id;
    params_node* params;
    compound_stmt_node* body;
    
    fun_decl_node(int line, char* s, params_node* p, compound_stmt_node* b);
    ~fun_decl_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct param_list_node : params_node
{
    param_list_node* head;
    param_node* tail;
    
    param_list_node(int line, param_list_node* h, param_node* t);
    ~param_list_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct simple_param_node : param_node
{
    type_spec_node* type;
    char* id;

    simple_param_node(int line, type_spec_node* t, char* s);
    ~simple_param_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct arr_param_node : param_node
{
    type_spec_node* type;
    char* id;

    arr_param_node(int line, type_spec_node* t, char* s);
    ~arr_param_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};


struct loc_decl_node : node
{
    loc_decl_node* head;
    var_decl_node* tail;

    loc_decl_node(int line, loc_decl_node* h, var_decl_node* t);
    ~loc_decl_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct stmt_list_node : node
{
    stmt_list_node* head;
    stmt_node* tail;

    stmt_list_node(int line, stmt_list_node* h, stmt_node* t);
    ~stmt_list_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct expr_stmt_node : stmt_node
{
    expr_node* expr;

    expr_stmt_node(int line, expr_node* e);
    ~expr_stmt_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct select_stmt_node : stmt_node
{
    expr_node* expr;
    stmt_node* ifstmt;
    stmt_node* elsestmt;

    select_stmt_node(int line, expr_node* e, stmt_node* ifs, stmt_node* els);
    ~select_stmt_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct iter_stmt_node : stmt_node
{
    expr_node* expr;
    stmt_node* body;

    iter_stmt_node(int line, expr_node* e, stmt_node* b);
    ~iter_stmt_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct return_stmt_node : stmt_node
{
    expr_node* expr;

    return_stmt_node(int line, expr_node* e);
    ~return_stmt_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct assign_expr_node : expr_node
{
    var_node* var;
    expr_node* expr;

    assign_expr_node(int line, var_node* v, expr_node* e);
    ~assign_expr_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct simple_var_node : var_node
{
    char* id;

    simple_var_node(int line, char* s);
    ~simple_var_node() { free(id); }
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct arr_var_node : var_node
{
    char* id;
    expr_node* expr;

    arr_var_node(int line, char* s, expr_node* e);
    ~arr_var_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct compound_simple_expr_node : simple_expr_node
{
    add_expr_node* expr1;
    relop_node* rel;
    add_expr_node* expr2;

    compound_simple_expr_node(int line, add_expr_node* e1, relop_node* r,
                              add_expr_node* e2);
    ~compound_simple_expr_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

enum relop_t { REL_LESS, REL_LESS_EQ, REL_GREATER, REL_GREATER_EQ,
               REL_EQUALS, REL_NEQUALS };
               
struct relop_node : node
{
    relop_t op;

    relop_node(int line, relop_t o);
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct compound_add_expr_node : add_expr_node
{
    add_expr_node* expr1;
    addop_node* op;
    term_node* expr2;

    compound_add_expr_node(int line, add_expr_node* e1, addop_node* o, term_node* e2);
    ~compound_add_expr_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

enum addop_t { ADD_PLUS, ADD_MINUS };
struct addop_node : node
{
    addop_t op;

    addop_node(int line, addop_t o);
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct compound_term_node : term_node
{
    term_node* expr1;
    mulop_node* op;
    factor_node* expr2;

    compound_term_node(int line, term_node* e1, mulop_node* o, factor_node* e2);
    ~compound_term_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

enum mulop_t { MUL_TIMES, MUL_DIVIDE };
struct mulop_node : node
{
    mulop_t op;

    mulop_node(int line, mulop_t o);
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct paren_factor_node : factor_node
{
    expr_node* expr;

    paren_factor_node(int line, expr_node* e);
    ~paren_factor_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct var_factor_node : factor_node
{
    var_node* var;

    var_factor_node(int line, var_node* v);
    ~var_factor_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct num_factor_node : factor_node
{
    char* val;

    num_factor_node(int line, char* v, int t);
    ~num_factor_node() { free(val); }
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct call_node : factor_node
{
    char* id1;
    args_node* args;

    call_node(int line, char* s1, args_node* a);
    ~call_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct args_node : node
{
    arg_list_node* args;

    args_node(int line, arg_list_node* a);
    ~args_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct arg_list_node : node
{
    arg_list_node* head;
    expr_node* tail;

    arg_list_node(int line, arg_list_node* h, expr_node* t);
    ~arg_list_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct sync_node : module_node
{
    char* id;
    sync_decl_list_node* tail;

    sync_node(int line, char* s, sync_decl_list_node* t);
    ~sync_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct global_sync_node : module_node
{
    sync_decl_list_node* tail;

    global_sync_node(int line, sync_decl_list_node* t);
    ~global_sync_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct sync_init_node : node
{
    actor_params_node* head;
    compound_stmt_node* tail;

    sync_init_node(int line, actor_params_node* h, compound_stmt_node* t);
    ~sync_init_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct constraint_block_node : node
{
    constraint_list_node* head;

    constraint_block_node(int line, constraint_list_node* h);
    ~constraint_block_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct constraint_list_node : node
{
    constraint_list_node* head;
    constraint_node* tail;

    constraint_list_node(int line, constraint_list_node* h, constraint_node* t);
    ~constraint_list_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct constraint_node : node
{
    // 0 = MIN, 1 = MAX
    int type;
    char* actor1;
    char* method1;
    char* actor2;
    char* method2;
    expr_node* tail;

    constraint_node(int line, int ct, char* a1, char* m1,
                    char* a2, char* m2, expr_node* t);
                    
    ~constraint_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct trigger_block_node : node
{
    trigger_list_node* head;

    trigger_block_node(int line, trigger_list_node* h);
    ~trigger_block_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct trigger_list_node : node
{
    trigger_list_node* head;
    trigger_node* tail;

    trigger_list_node(int line, trigger_list_node* h, trigger_node* t);
    ~trigger_list_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct ed_trigger_node : trigger_node
{
    // 0 disable, 1 enable
    int type;
    char* actor;
    char* method;
    expr_node* expr;
    
    ed_trigger_node(int line, int t, char* a, char* m, expr_node* e);
    ~ed_trigger_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct action_trigger_node : trigger_node
{
    char* actor;
    char* method;
    compound_stmt_node* tail;
    
    action_trigger_node(int line, char* a, char* m, compound_stmt_node* c);
    ~action_trigger_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct actor_params_node : node
{
    actor_params_node* head;
    actor_param_node* tail;
        
    actor_params_node(int line, actor_params_node* h, actor_param_node* tail);
    ~actor_params_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct actor_param_node : node
{
    char* id1;
    char* id2;

    actor_param_node(int line, char* i1, char* i2);
    ~actor_param_node() { free(id1); free(id2); }
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct func_decls_node : node
{
    func_decls_node* head;
    fun_decl_node* tail;
  
    func_decls_node(int line, func_decls_node* h, fun_decl_node* t);
    ~func_decls_node();
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};

struct pass_node : module_node
{
    char* text;

    pass_node(int line, char* txt);
    ~pass_node() { free(text); }
    void accept(int depth, visitor& v) { v.visit(depth, this); }
};


#endif // #define AST_H
