/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "codevisitor.h"
#include "code_param_visitor.h"
#include "gsyncvisitor.h"


// Please refer to the YACC specification to get some sense of grammatical
// structure.


void codevisitor::visit(int d, node* n) {
    out << "PARSER ERROR: Trying to parse base node\n";
}

void codevisitor::visit(int d, program_node* n) {
    out << "#include <map>\n"
        << "#include <sstream>\n"
        << "#include <vector>\n"
        << "#include \"runtime/logger.h\"\n"
        << "#include \"runtime/postmaster.h\"\n"
        << "#include \"runtime/actor.h\"\n"
        << "#include \"runtime/messages.h\"\n"
        << "#include \"runtime/synchronizer.h\"\n"
        << "#include \"runtime/globalsync.h\"\n\n"
        << "\n"
        << "void rt_start(const actorref& a, const method& m, "
        << "unsigned short serverport=6127, const char* host=0, "
        << "unsigned short host_port=6127);\n"
        << "void rt_start("
        << "unsigned short serverport=6127, const char* host=0, "
        << "unsigned short host_port=6127);\n\n";

    n->modules->accept(d+1, *this);

    out << "void setup_run(postmaster& PostMaster)\n"
        << "{\n";

    for(std::list<actor_id>::iterator it = actor_list.begin();
        it != actor_list.end(); ++it)
    {
        out << indent(1) << "PostMaster.reg(\"" << it->first << "\", new "
            << it->second << "(PostMaster));\n";
    }
    for(std::list<sync_id>::iterator it = sync_list.begin();
        it != sync_list.end(); ++it)
    {
        out << indent(1) << "PostMaster.reg(\"" << it->first << "\", new "
            << it->second << "(PostMaster));\n";
    }
    out << "std::stringstream ss;\n"
        << "ss << PostMaster.get_nid() << \"global\";\n";
    for(std::list<sync_id>::iterator it = gsync_list.begin();
        it != gsync_list.end(); ++it)
    {
        out << indent(1) << "PostMaster.reg(ss.str(), new "
            << it->second << "(ss.str(), PostMaster));\n";
    }
    out << "}\n\n";


    out << "\nvoid rt_start("
        << "unsigned short serverport, const char* host, "
        << "unsigned short host_port)\n"
        << "{\n"
        << indent(1) << "postmaster PostMaster(host, host_port, "
        << "serverport);\n"
        << indent(1) << "setup_run(PostMaster);\n"
        << indent(1) << "PostMaster.run(host != 0);\n"
        << "}\n\n";

    out << "void rt_start(const actorref& a, const method& m, "
        << "unsigned short serverport, const char* host, "
        << "unsigned short host_port)\n"
        << "{\n"
        << indent(1) << "postmaster PostMaster(host, host_port, "
        << "serverport);\n"
        << indent(1) << "setup_run(PostMaster);\n"
        << indent(1) << "PostMaster.run(a, m, host != 0);\n"
        << "}\n\n";
}

void codevisitor::visit(int d, decl_node* n) {
    ;
}

void codevisitor::visit(int d, var_decl_node* n) {
    ;
}

void codevisitor::visit(int d, simple_var_decl_node* n) {
    out << indent(d);
    n->type->accept(d+1, *this);
    out << n->id << ";\n";
}

// Arrays are implemented as vectors
void codevisitor::visit(int d, arr_var_decl_node* n) {
    /*out << indent(d) << "std::vector<";
    n->type->accept(d+1, *this);
    //out << "> " << n->id << "(" << n->elems << ");\n";
    out << "> " << n->id <<";\n";*/

    //changed from being a vector to an actual array

    n->type->accept(d+1, *this);
    out <<" " << n->id <<"[" <<n->elems <<"];\n";
}

void codevisitor::visit(int d, type_spec_node* n) {
    switch(n->type)
    {
        case TK_VOID:
            out << "void ";
            cur_params << "void ";
            break;
        case TK_INT:
            out << "int ";
            cur_params << "int ";
            break;
        case TK_FLOAT:
            out << "float ";
            cur_params << "float ";
            break;
        case TK_STRING:
            out << "std::string ";
            cur_params << "string ";
            break;
        default:
            break;
    }
}

void codevisitor::visit(int d, exception_block_node* n)
{
    isInException = true;
    out << indent(d) << "int exception_handler(const timing_exception& te)\n"
        << indent(d) << "{\n";
    n->body->accept(d+1, *this);
    out << indent(d+1) << "return TERMINATE;\n"
        << indent(d) << "}\n\n";
    isInException = false;
}

// Generate method.
// Note that for each method, two functions are actually generated: An
// internal function that the actor can call freely from other methods, and
// an external function which the execute() function invokes with a params
// object.
void codevisitor::visit(int d, fun_decl_node* n) {
    // Track what method we are in
    cur_methods.push_back(std::string(n->id));

    graphout << "<method>\n" << cur_entity << "\n" << n->id << "\n";

    // Print signature
    out << indent(d);
    out << "rttime_t " << n->id << "(";
    cur_params.str("");
    cur_params.clear();
    if(n->params) n->params->accept(d+1, *this);
    // Save current parameters; messed up by every type specifier ...
    std::string paramstr = cur_params.str();
    out << ")\n"
        << indent(d) << "{\n";

    // Print body
    n->body->accept(d+2, *this);
    // Notify call -- note extra set of braces (compound_stmt is a block
    // nested in the body)
    out << indent(d+1) << "PostMaster.update_local_run(name,\"" << n->id << "\");\n"
        << indent(d+1) << "rttime_t t = rt_gettime();\n";

    out << indent(d+1) << "notify(std::string(\"" << n->id << "\"), t);\n"
        << indent(d+1) << "return t;\n"
        << indent(d) << "}\n\n";

    // Now generate an overloaded function that accepts a params object and
    // decodes it into parameters
    out << indent(d);
    out << "rttime_t " << n->id << "(params& p)\n" << indent(d) << "{\n";


    // Type is stored in cur_params as "int float char vector int string",
    // where "vector" -precedes- the type of array elements
    int cur_var = 0;
    cur_params.str(paramstr);
    std::string buf;
    while((cur_params >> buf))
    {
        // "3 5 8 12 432 3 4 $ "
        if(buf == "vector")
        {
            cur_params >> buf;
            out << indent(2) << "std::vector<" << buf << " > var" << cur_var++ << ";\n";
        }
        else if(buf == "string")
        {
            out << indent(2) << "std::string var" << cur_var++ << ";\n";
        }
        else
            out << indent(2) << buf << " var" << cur_var++ << ";\n";
    }

    // Decode by reading from params, stringstream-like
    cur_params.str(paramstr);
    if(cur_var)
        out << indent(2) << "p";
    for(int i = 0; i < cur_var; ++i)
    {
        out << " >> var" << i;
    }
    if(cur_var)
        out << ";\n";

    // Generate call of overloaded method
    out << indent(2) << "return " << n->id << "(";
    for(int i = 0; i < cur_var; ++i)
        out << "var" << i << (i+1 < cur_var ? ", " : "");
    out << ");\n";

    // Close method
    out << indent(1) << "}\n\n";
    graphout << "</method>\n";
}

void codevisitor::visit(int d, params_node* n) {
    ;
}

void codevisitor::visit(int d, param_list_node* n) {
    if(n->head) {
        n->head->accept(d, *this);
        out << ", ";
    }
    n->tail->accept(d, *this);
}

void codevisitor::visit(int d, param_node* n) {
    ;
}

void codevisitor::visit(int d, simple_param_node* n) {
    n->type->accept(d+1, *this);
    out << n->id;
}

void codevisitor::visit(int d, arr_param_node* n) {
    cur_params << "vector ";
    out << "std::vector<";
    n->type->accept(d+1, *this);
    out << " >& " << n->id;
}

void codevisitor::visit(int d, compound_stmt_node* n) {
    out << indent(d-1) << "{\n";
    if(n->decls) n->decls->accept(d, *this);
    if(n->body) n->body->accept(d, *this);
    out << indent(d-1) << "}\n";
}

void codevisitor::visit(int d, loc_decl_node* n) {
    if(n->head) n->head->accept(d, *this);
    n->tail->accept(d, *this);
}

void codevisitor::visit(int d, stmt_list_node* n) {
    if(!n) return;
    if(n->head) n->head->accept(d, *this);
    if(n->tail) n->tail->accept(d, *this);
}

void codevisitor::visit(int d, stmt_node* n) {
    ;
}

void codevisitor::visit(int d, expr_stmt_node* n) {
    out << indent(d);
    if(n->expr) n->expr->accept(d+1, *this);
    out << ";\n";
}

void codevisitor::visit(int d, select_stmt_node* n) {
    std::stringstream ss, junk;
    codevisitor v(ss, junk);
    n->expr->accept(d+1, v);

    std::stringstream tt;
    tt << "if " << ss.str() << "\n";
    // Track scopes
    scope_exprs.push_back(tt.str());
    tt.str("");

    out << indent(d) << "if (";
    n->expr->accept(d+1, *this);
    out << ")\n" << indent(d) << "{\n";
    n->ifstmt->accept(d+1, *this);
    out << indent(d) << "}\n";

    // Exiting scpoe
    scope_exprs.pop_back();

    if(n->elsestmt) {
        tt << "else !(" << ss.str() << ")\n";
        scope_exprs.push_back(tt.str());

        out << indent(d) << "else\n" << indent(d) << "{\n";
        n->elsestmt->accept(d+1, *this);
        out << indent(d) << "}\n";

        scope_exprs.pop_back();
    }
}

void codevisitor::visit(int d, iter_stmt_node* n) {
    std::stringstream ss, junk;
    ss << "while ";
    codevisitor v(ss, junk);
    n->expr->accept(d+1, v);
    ss << "\n";
    // Track scopes
    scope_exprs.push_back(ss.str());

    out << indent(d) << "while (";
    n->expr->accept(d+1, *this);
    out << ")\n";
    n->body->accept(d+1, *this);

    // Exiting scope
    scope_exprs.pop_back();
}

void codevisitor::visit(int d, return_stmt_node* n)
{
    if (isInException)
    {
        out << indent(d) << "return ";
        n->expr->accept(d, *this);
        out << ";\n";
    }
    else
    {
        throw "Return statement not supported here.";
    }
}

void codevisitor::visit(int d, expr_node* n) {
    ;
}

void codevisitor::visit(int d, assign_expr_node* n) {
    n->var->accept(d+1, *this);
    out << " = ";
    n->expr->accept(d+1, *this);
}

void codevisitor::visit(int d, var_node* n) {
    ;
}

void codevisitor::visit(int d, simple_var_node* n) {
    out << n->id;
}

void codevisitor::visit(int d, arr_var_node* n) {
    out << n->id << "[";
    n->expr->accept(d+1, *this);
    out << "]";
}

void codevisitor::visit(int d, simple_expr_node* n) {
    ;
}

void codevisitor::visit(int d, compound_simple_expr_node* n) {
    n->expr1->accept(d+1, *this);
    n->rel->accept(d+1, *this);
    n->expr2->accept(d+1, *this);
}

void codevisitor::visit(int d, relop_node* n) {
    switch(n->op)
    {
        case REL_LESS:
            out << " < ";
            break;
        case REL_LESS_EQ:
            out << " <= ";
            break;
        case REL_GREATER:
            out << " > ";
            break;
        case REL_GREATER_EQ:
            out << " >= ";
            break;
        case REL_EQUALS:
            out << " == ";
            break;
        case REL_NEQUALS:
            out << " != ";
            break;
        default:
            break;
    }
}

void codevisitor::visit(int d, add_expr_node* n) {
    ;
}

void codevisitor::visit(int d, compound_add_expr_node* n) {
    n->expr1->accept(d+1, *this);
    n->op->accept(d+1, *this);
    n->expr2->accept(d+1, *this);
}

void codevisitor::visit(int d, addop_node* n) {
    switch(n->op)
    {
        case ADD_PLUS:
            out << " + ";
            break;
        case ADD_MINUS:
            out << " - ";
            break;
        default:
            break;
    }
}

void codevisitor::visit(int d, term_node* n) {
    ;
}

void codevisitor::visit(int d, compound_term_node* n) {
    n->expr1->accept(d+1, *this);
    n->op->accept(d+1, *this);
    n->expr2->accept(d+1, *this);
}

void codevisitor::visit(int d, mulop_node* n) {
    switch(n->op)
    {
        case MUL_TIMES:
            out << " * ";
            break;
        case MUL_DIVIDE:
            out << " / ";
            break;
        default:
            break;
    }
}

void codevisitor::visit(int d, factor_node* n) {
    ;
}

void codevisitor::visit(int d, paren_factor_node* n) {
    out << "(";
    n->expr->accept(d+1, *this);
    out << ")";
}

void codevisitor::visit(int d, var_factor_node* n) {
    n->var->accept(d+1, *this);
}

void codevisitor::visit(int d, num_factor_node* n) {
    out << n->val;
}

void codevisitor::visit(int d, sync_call_stmt_node* n) {
    // Print all conditions for this call
    graphout << scope_exprs.size() << "\n";
    for(std::vector<std::string>::iterator it = scope_exprs.begin();
        it != scope_exprs.end(); ++it)
    {
        graphout << *it;
    }

    out << indent(d) << "{\n";
    out << indent(d+1) << "message m(\"" << cur_entity << "\", \""
        << cur_methods.back() << "\", \"" << n->id1 << "\", \"" << n->id2
        << "\", params());\n";
    out << indent(d+1) << "m.synchronous = true;\n";
    out << indent(d+1) << "rt_log << \"created synchronous message \" << m.serial "
                        << "<< \"\\n\" << m;\n";
    out << indent(d+1) << "PostMaster.sendsyncmsg(m);\n";
    out << indent(d) << "}\n";
    graphout << n->id1 << "\n" << n->id2 << "\n";
}

void codevisitor::visit(int d, call_stmt_node* n) {
    // Print all conditions for this call
    graphout << scope_exprs.size() << "\n";
    for(std::vector<std::string>::iterator it = scope_exprs.begin();
        it != scope_exprs.end(); ++it)
    {
        graphout << *it;
    }

    out << indent(d) << "{\n";
    out << indent(d) << "params p;\n";
    code_param_visitor param_visitor(out, graphout);
    if(n->args) n->args->accept(d+2, param_visitor);
    out << indent(d) << "message m(\"" << cur_entity << "\", \""
        << cur_methods.back() << "\", \"" << n->id1 << "\", \"" << n->id2
        << "\", p);\n";
    out << indent(d) << "rt_log << \"created message \" << m.serial "
                        << "<< \"\\n\" << m;\n";
    out << indent(d) << "PostMaster.sendmsg(m);\n";
    out << indent(d-1) << "}\n";
    graphout << n->id1 << "\n" << n->id2 << "\n";
}

void codevisitor::visit(int d, call_node* n) {
    out << n->id1 << "(";
    if(n->args) n->args->accept(d+1, *this);
    out << ")";
}

void codevisitor::visit(int d, args_node* n) {
    n->args->accept(d+1, *this);
}

void codevisitor::visit(int d, arg_list_node* n) {
    if(n->head) {
        n->head->accept(d, *this);
        out << ", ";
    }
    n->tail->accept(d, *this);
}


// Generate the actor class istelf
void codevisitor::visit(int d, actor_node* n) {
    cur_entity = std::string(n->id);
    std::string cur_classname = "actor_" + cur_entity + "_t";
    actor_list.push_back(actor_id(cur_entity, cur_classname));
    cur_methods.clear();
    out << "class " << cur_classname << " : public actor\n{\n"
        << indent(1)
        << "typedef rttime_t (" << cur_classname << "::*local_method)"
        << "(params&);\n"
        // Track run timestamps of methods
        << indent(1) << "struct method_t" << indent(1) << "{\n"
        << indent(2) << "local_method m;\n"
        << indent(2) << "method_t() { } // Never called!\n"
        << indent(2) << "method_t(local_method mt)\n"
        << indent(2) << ": m(mt)\n"
        << indent(2) << "{ }\n"
        << indent(1) << "};\n"
        << indent(1) << "std::map<const method, method_t> methods;\n\n";
    n->tail->accept(1, *this);

    // Dispatch message to appropriate member function
    out << "\n" << indent(1) << "void execute(message& m)\n"
        << indent(1) << "{\n"
        << indent(2) << "rt_current = m.must_run_before;\n"
        << indent(2) << "rt_log << name << \" executing \" << m.callee << \"\\n\";\n"
        << indent(2) << "rttime_t t = (this->*(methods[m.callee].m))(m.m_params);\n"
        << indent(2) << "m.timestamp.executed = t;\n"
        << indent(2) << "PostMaster.sm_ack_msg(m.sync, m);\n"
        << indent(2) << "if(m.synchronous)\n"
        << indent(2) << "{\n"
        << indent(3) << "PostMaster.recvsyncmsg(m);\n"
        << indent(2) << "}\n"
        << indent(2) << "rt_current = rt_inf;\n"
        << indent(1) << "}\n";

    out << "\n" << indent(1) << "~" << cur_classname << "()\n"
        << indent(1) << "{\n"
        << indent(2) << "rt_log << \"Killing\" << name << \"\\n\";\n"
        << indent(2) << "end_polling();\n"
        << indent(1) << "}\n\n";

    out << "};\n";
}

void codevisitor::visit(int d, decl_list_node* n) {
    if(n){
        if(n->vars)
            n->vars->accept(d, *this);
        out << "\n";
        if(n->methods)
            n->methods->accept(d, *this);

        out << "\npublic:\n";
        n->init->accept(d, *this);
    }
}

void codevisitor::visit(int d, func_decls_node* n) {
    if(n->head) n->head->accept(d, *this);
    n->tail->accept(d, *this);
}


void codevisitor::visit(int d, init_node* n) {
    out << "\n" << indent(1) << "actor_" << cur_entity << "_t(postmaster& p)"
        << " : actor(\""
        << cur_entity << "\",p)\n" << indent(1) << "{\n"
        << indent(2) << "std::list<method> method_list;"
        << indent(2) << "// Add methods to lookup map\n";
    for(std::list<std::string>::iterator it = cur_methods.begin();
        it != cur_methods.end(); ++it)
    {
        out << indent(2) << "methods[\"" << *it << "\"] = ";
        out << "method_t((local_method) &actor_" << cur_entity << "_t::" << *it << ");\n";
        out << indent(2) << "method_list.push_back(\"" << *it << "\");\n";
    }

    out << "\n" << indent(2) << "// init block\n";
    n->head->accept(3, *this);

    out << indent(2) << "// Register with PostMaster\n"
        << indent(2) << "PostMaster.reg_methods(\"" << cur_entity << "\", method_list);\n"
        << indent(2) << "start_polling();\n";
    out << indent(1) << "}\n";
}

void codevisitor::visit(int d, var_decls_node* n) {
    if(n->head) n->head->accept(d, *this);
    n->tail->accept(d, *this);
}

void codevisitor::visit(int d, global_sync_node* n)
{
    gsyncvisitor v(out, graphout);
    gsync_list.push_back(sync_id(cur_entity, "sync_global_t"));
    n->tail->accept(d, v);
}

void codevisitor::visit(int d, sync_node* n) {
    cur_entity = n->id;
    sync_list.push_back(sync_id(cur_entity, "sync_" + cur_entity + "_t"));
    graphout << "<sync>\n" << cur_entity << "\n";
    if(cur_entity == "global")
        out << "class sync_" << n->id << "_t : public globalsync\n{\n";
    else
        out << "class sync_" << n->id << "_t : public synchronizer\n{\n";
    n->tail->accept(d+1, *this);
    out << "};\n";
    graphout << "</sync>\n";
}

void codevisitor::visit(int d, sync_decl_list_node* n) {
    if(n->vars)
        n->vars->accept(1, *this);
    out << "\npublic:\n";
    n->constraints->accept(1, *this);
    out << "\n";
    n->triggers->accept(1, *this);
    out << "\n";
    n->init->accept(1, *this);
    out << "\n";
    n->exception->accept(1, *this);
}

void codevisitor::visit(int d, constraint_block_node* n) {
    out << indent(1) << "poll_result poll_msg(message& msg)\n"
        << indent(1) << "{\n";
    if(n->head) n->head->accept(2, *this);
    out << "\n";
    out << indent(2) << "rttime_t now = rt_gettime();\n"
        << indent(2) << "if(now >= msg.must_run_before)\n"
        << indent(3) << "throw timing_exception(msg);\n";
    out << indent(2) << "else if(now > msg.must_run_after)\n"
        << indent(3) << "return REMOVE;\n";
    out << indent(2) << "else\n"
        << indent(3) << "return DEFAULT;\n";
    out << indent(1) << "}\n\n";
}

void codevisitor::visit(int d, constraint_list_node* n) {
    if(n->head) n->head->accept(d, *this);
    n->tail->accept(d, *this);
}


void codevisitor::visit(int d, constraint_node* n)
{
    if(n->type == 0) // MIN
    {
        // Set time constraints
        out << indent(2) << "if(msg.receiver == \"" << n->actor2
            << "\" and msg.callee == \"" << n->method2 << "\")\n"
            << indent(2) << "{\n"
            << indent(3) << "rttime_t min = PostMaster.am_lastrun(\""
            << n->actor1 << "\", \"" << n->method1 << "\") + rttime_t(0, ";

        graphout << "min\n" << n->actor1 << "\n" << n->method1 << "\n"
                 << n->actor2 << "\n" << n->method2 << "\n";
        codevisitor v(graphout);
        n->tail->accept(2, *this);
        n->tail->accept(0, v);
        graphout << "\n";

        out << ");\n";
        out << indent(3) << "if(msg.must_run_after < min)\n" << indent(3)
            << "{\n" << indent(4) << "msg.must_run_after = min;\n"
            << indent(4) << "rt_log << msg;\n" << indent(3) << "}\n"
            << indent(2) << "}\n";
    }
    else // MAX
    {
        // Set time constraints
        out << indent(2) << "if(msg.receiver == \"" << n->actor2
            << "\" and msg.callee == \"" << n->method2 << "\")\n"
            << indent(2) << "{\n"
            << indent(3) << "rttime_t max = PostMaster.am_lastrun(\""
            << n->actor1 << "\", \"" << n->method1 << "\") + rttime_t(0, ";

        graphout << "max\n" << n->actor1 << "\n" << n->method1 << "\n"
                 << n->actor2 << "\n" << n->method2 << "\n";
        codevisitor v(graphout);
        n->tail->accept(2, *this);
        n->tail->accept(0, v);
        graphout << "\n";

        out << ");\n";
        out << indent(3) << "if(msg.must_run_before > max)\n" << indent(3)
            << "{\n" << indent(4) << "msg.must_run_before = max;\n"
            << indent(4) << "rt_log << msg;\n" << indent(3) << "}\n"
            << indent(2) << "}\n";
    }
}

void codevisitor::visit(int d, trigger_block_node* n) {
    if(n->head) n->head->accept(d+1, *this);

    if(cur_entity == "global")
    {
        out << indent(1) << "void notify_action(const actorref& a, const method& m, const rttime_t& t)\n"
            << indent(1) << "{\n"
            << indent(2) << "t_notify_action(a, m, t);\n"
            << indent(2) << "c_notify_action(a, m, t);\n"
            << indent(1) << "}\n\n";

        out << indent(1) << "void t_notify_action(const actorref& a, const method& m, const rttime_t& t)\n";
    }
    else
        out << indent(1) << "void notify_action(const actorref& a, const method& m, const rttime_t& t)\n";

    out << indent(1) << "{\n"
        << indent(2) << "rt_log << name << \" rec. execution notification from \" << a << \".\" << m << \" at \" << t << \"\\n\";\n"
        << action_triggers.str()
        << indent(1) << "}\n\n";

    out << indent(1) << "void poll_triggers()\n"
        << indent(1) << "{\n"
        << ed_triggers.str()
        << indent(1) << "}\n\n";

    ed_triggers.str("");
    action_triggers.str("");
}

void codevisitor::visit(int d, trigger_list_node* n) {
    if(n->head) n->head->accept(d, *this);
    n->tail->accept(d, *this);
}

void codevisitor::visit(int d, trigger_node* n) {
    ;
}

void codevisitor::visit(int d, ed_trigger_node* n)
{
    ed_triggers << indent(2) << "if(";
    codevisitor v(ed_triggers);
    n->expr->accept(d+1, v);
    ed_triggers << ")\n"
                << indent(3) << "if(" << n->type << "== 0) // Disable\n"
                << indent(4) << "PostMaster.am_disable(\"" << n->actor
                             << "\",\"" << n->method << "\");\n"
                << indent(3) << "else // Enable\n"
                << indent(4) << "PostMaster.am_enable(\"" << n->actor
                             << "\",\"" << n->method << "\");\n";
}

void codevisitor::visit(int d, action_trigger_node* n) {
    action_triggers << indent(2) << "if(\"" << n->actor << "\" == a and \""
                    << n->method << "\" == m)\n";
    codevisitor v(action_triggers);
    n->tail->accept(d+1, v);
}

void codevisitor::visit(int d, actor_params_node* n) {
    if(n->head) n->head->accept(d, *this);
    n->tail->accept(d+1, *this);
}

void codevisitor::visit(int d, actor_param_node* n) {
    // If it's just an actor (not synchronized), just register ourselves as a
    // listener
    if(!n->id2)
        out << indent(2) << "PostMaster.am_register_listener(\"" << n->id1
            << "\", name);\n";
    // Otherwise we're listening to both and synchronizing the combination
    else
    {
        out << indent(2) << "PostMaster.am_register_listener(\"" << n->id1
            << "\", name);\n";
        out << indent(2) << "PostMaster.am_register_listener(\"" << n->id2
            << "\", name);\n";

        out << indent(2) << "PostMaster.reg(\"" << cur_entity << "\", \""
            << n->id1 << "\", \"" << n->id2 << "\");\n";
    }
}


void codevisitor::visit(int d, sync_init_node* n) {
    if(cur_entity == "global")
    {
        out << indent(1) << "sync_" << cur_entity
            << "_t(const syncref& name, postmaster& p)"
            << " : globalsync(name, p)\n";
    }
    else
    {
        out << indent(1) << "sync_" << cur_entity << "_t(postmaster& p)"
            << " : synchronizer(\"" << cur_entity << "\",p)\n";
    }
    out << indent(1) << "{\n"
        << indent(2) << "//PostMaster.reg(\"" << cur_entity << "\", this);\n";
    n->head->accept(2, *this);
    n->tail->accept(3, *this);
    out << indent(2) << "start_polling();\n";
    out << indent(1) << "}\n\n";

    out << indent(1) << "~sync_" << cur_entity << "_t()\n"
        << indent(1) << "{\n"
        << indent(2) << "rt_log << \"Killing\" << name << \"\\n\";\n"
        << indent(2) << "end_polling();\n"
        << indent(1) << "}\n\n";
}

void codevisitor::visit(int d, module_list_node* n) {
    if(n->head) n->head->accept(d, *this);
    n->tail->accept(d, *this);
}

void codevisitor::visit(int d, module_node* n) {
    ;
}

void codevisitor::visit(int d, pass_node* n) {
    out << n->text;
}
