/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef CODEVISITOR_H
#define CODEVISITOR_H

#include "ast.h"
#include "visitor.h"
#include "../../runtime/messages.h"

#include <iostream>
#include <list>
#include <string>
#include <sstream>
#include <vector>

class gsyncvisitor;

// This is the pivotal Visitor class that handles all but some highly
// specialised code generation. Having evolved organically from an
// originally clean design, it now holds a certain amount of states for
// specialised situations.
class codevisitor : public visitor
{
    // Code output
    std::ostream& out;
    // Sequence graph output
    std::ostream& graphout;

    // Simple list of all actors
    typedef std::pair<actorref, std::string> actor_id;
    std::list<actor_id> actor_list;
    // List of synchronizers
    typedef std::pair<syncref, std::string> sync_id;
    std::list<sync_id> sync_list;
    // List of global synchronizers (should be 0 or 1)
    std::list<sync_id> gsync_list;

    // Which entity (actor/synchronizer) are we currently generating
    // code for?
    std::string cur_entity;
    // Parameters of the method we are generating, if applicable
    std::stringstream cur_params;
    // Methods of current actor, if applicable
    std::list<std::string> cur_methods;
    // Store triggers
    std::stringstream ed_triggers;
    std::stringstream action_triggers;
    // True if in exception block of a synchronizer
    bool isInException;
    // Stack of scopes. This is used for graph generation to track
    // conditionals in graph edges; each scope represents the
    // type of conditional statement and the conditional expression
    // of the current scope. See graphmaker files for details.
    std::vector<std::string> scope_exprs;

    codevisitor(std::ostream& outstream)
    : out(outstream), graphout(outstream), isInException(false) { }
    friend class gsyncvisitor;
public:
    codevisitor(std::ostream& outstream, std::ostream& graph)
    : out(outstream), graphout(graph), isInException(false) { }

    void visit(int d, node* n);
    void visit(int d, program_node* n);
    void visit(int d, decl_list_node* n);
    void visit(int d, decl_node* n);
    void visit(int d, var_decl_node* n);
    void visit(int d, simple_var_decl_node* n);
    void visit(int d, arr_var_decl_node* n);
    void visit(int d, type_spec_node* n);
    void visit(int d, exception_block_node* n);
    void visit(int d, fun_decl_node* n);
    void visit(int d, params_node* n);
    void visit(int d, param_list_node* n);
    void visit(int d, param_node* n);
    void visit(int d, simple_param_node* n);
    void visit(int d, arr_param_node* n);
    void visit(int d, compound_stmt_node* n);
    void visit(int d, loc_decl_node* n);
    void visit(int d, stmt_list_node* n);
    void visit(int d, stmt_node* n);
    void visit(int d, expr_stmt_node* n);
    void visit(int d, select_stmt_node* n);
    void visit(int d, iter_stmt_node* n);
    void visit(int d, return_stmt_node* n);
    void visit(int d, expr_node* n);
    void visit(int d, assign_expr_node* n);
    void visit(int d, var_node* n);
    void visit(int d, simple_var_node* n);
    void visit(int d, arr_var_node* n);
    void visit(int d, simple_expr_node* n);
    void visit(int d, compound_simple_expr_node* n);
    void visit(int d, relop_node* n);
    void visit(int d, add_expr_node* n);
    void visit(int d, compound_add_expr_node* n);
    void visit(int d, addop_node* n);
    void visit(int d, term_node* n);
    void visit(int d, compound_term_node* n);
    void visit(int d, mulop_node* n);
    void visit(int d, factor_node* n);
    void visit(int d, paren_factor_node* n);
    void visit(int d, var_factor_node* n);
    void visit(int d, num_factor_node* n);
    void visit(int d, call_node* n);
    void visit(int d, args_node* n);
    void visit(int d, arg_list_node* n);

    void visit(int d, actor_node* n);
    void visit(int d, sync_node* n);
    void visit(int d, global_sync_node* n);
    void visit(int d, init_node* n);
    void visit(int d, sync_init_node* n);
    void visit(int d, sync_decl_list_node* n);
    void visit(int d, constraint_block_node* n);
    void visit(int d, constraint_list_node* n);
    void visit(int d, constraint_node* n);
    void visit(int d, trigger_block_node* n);
    void visit(int d, trigger_list_node* n);
    void visit(int d, trigger_node* n);
    void visit(int d, ed_trigger_node* n);
    void visit(int d, action_trigger_node* n);
    void visit(int d, actor_params_node* n);
    void visit(int d, actor_param_node* n);
    void visit(int d, var_decls_node* n);
    void visit(int d, func_decls_node* n);
    void visit(int d, module_list_node* n);
    void visit(int d, module_node* n);
    void visit(int d, pass_node* n);

    void visit(int d, call_stmt_node* n);
    void visit(int d, sync_call_stmt_node* n);
};

#endif // #define CODEVISITOR_H
