/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "scopevisitor.h"
#include "codevisitor.h"
#include "ast.h"

#include <fstream>
#include <iostream>
#include <sstream>


std::string type_to_str(const keyword_t t)
{
    switch(t)
    {
        case TK_VOID:
            return "void";
        case TK_INT:
            return "int";
        case TK_FLOAT:
            return "float";
        case TK_STRING:
            return "string";
        default:
            return "";
    }
}


// This function takes two types and returns whatever type it widens to.
// For example, an expression involving an int and a float results in a
// float; with two ints, an int. If either type is UNKNOWN the result is
// UNKNOWN; if either is invalid, then the result is also INVALID.
int check_types(const int ls_type, const int rs_type)
{
    static const int I = expr_node::EX_INT;
    static const int F = expr_node::EX_FLOAT;
    static const int INV = expr_node::EX_INV;

    if ((ls_type & 0xF) == (rs_type & 0xF))
        return ls_type;
    if ((ls_type == I && rs_type == F) || (ls_type == F && rs_type == I))
        return F;
    if(ls_type == expr_node::EX_UNKNOWN)
        return rs_type;
    if(rs_type == expr_node::EX_UNKNOWN)
        return ls_type;
    return INV;
}


// Verify that the symbolic type t corresponds with the string representation.
// Note the somewhat arcane syntax for checking array status:
//    !!(t&expr_node::EX_ARRAY) == arr
// What is being done is this: The array bit from t is extracted via a bitwise
// AND, double-banged to get a boolean value (0 or 1), then compared against
// the requested type; on a match, the expression is true, else it is false.
bool match(int t, const std::string& typ, bool arr=false)
{
    switch(t & !expr_node::EX_ARRAY)
    {
        case expr_node::EX_INV:
            return false;
        case expr_node::EX_INT:
            return typ == "int" && (!!(t&expr_node::EX_ARRAY) == arr);
        case expr_node::EX_FLOAT:
            return typ == "float" && (!!(t&expr_node::EX_ARRAY) == arr);
        case expr_node::EX_STRING:
            return typ == "string" && (!!(t&expr_node::EX_ARRAY) == arr);
        case expr_node::EX_UNKNOWN:
            return true;
    }
}

boost::shared_ptr<scopevisitor::declarable> scopevisitor::find_symbol
    (const std::string& name)
{
    using namespace std;
    for(list<table>::reverse_iterator ri = symbol_table.rbegin();
        ri != symbol_table.rend(); ++ri)
    {
        table::iterator it = ri->find(name);
        if(it != ri->end())
            return it->second;
    }
    return boost::shared_ptr<declarable>((declarable*)0);
}

void scopevisitor::visit(int d, program_node* n)
{
    enter_scope();
    n->modules->accept(d,*this);
    leave_scope();
}

void scopevisitor::visit(int d, decl_list_node* n)
{
    if(n){
        if(n->vars)
            n->vars->accept(d, *this);
        n->init->accept(d, *this);
        if(n->methods)
            n->methods->accept(d, *this);
    }
}

// Check variable declarations against multiple declarations
void scopevisitor::visit(int d, simple_var_decl_node* n)
{
    table& map = symbol_table.back();

    std::string type = type_to_str(n->type->type);
    if(map.find(n->id) != map.end())
    {
        using namespace std;
        cout << n->line << ": " << type << " " << n->id
             << " was already declared in this scope" << endl;
        m_good = false;
    }
    else
        map[n->id] = boost::shared_ptr<declarable>(new var(type, n->id));
}

// Check variable declarations against multiple declarations
void scopevisitor::visit(int d, arr_var_decl_node* n)
{
    table& map = symbol_table.back();

    std::string type = type_to_str(n->type->type);
    if(map.find(n->id) != map.end())
    {
        using namespace std;
        cout << n->line << ": " << type << " " << n->id << "[]"
             << " was already declared in this scope" << endl;
        m_good = false;
    }
    else
        map[n->id] = boost::shared_ptr<declarable>(new var(type, n->id, true));
}

void scopevisitor::visit(int d, exception_block_node* n)
{
    n->body->accept(d+1,*this);
}

// Enter a function definition: Set up scopes and symbol tables
void scopevisitor::visit(int d, fun_decl_node* n)
{
    table& map = symbol_table.back();

    if(map.find(n->id) != map.end())
    {
        using namespace std;
        cout << n->line << ": method " << n->id
             << " was already declared in this scope\n"
             << "(overloading is not allowed)" << endl;
        m_good = false;
    }
    else
    {
        if(n->params)
            n->params->accept(d, *this);
        map[n->id] = boost::shared_ptr<declarable>(new func(n->id,cur_params));
    }
    enter_scope();
    table& tab = symbol_table.back();
    for(paramlist::iterator it = cur_params.begin(); it != cur_params.end();
        ++it)
    {
        tab[it->name] = boost::shared_ptr<declarable>(new var(*it));
    }
    n->body->accept(d+1,*this);
    leave_scope();
    cur_params.clear();
}

void scopevisitor::visit(int d, param_list_node* n)
{
    if(n->head)
        n->head->accept(d, *this);
    n->tail->accept(d, *this);
}

void scopevisitor::visit(int d, simple_param_node* n)
{
    std::string type = type_to_str(n->type->type);

    cur_params.push_back(var(type, n->id));
}

void scopevisitor::visit(int d, arr_param_node* n)
{
    std::string type = type_to_str(n->type->type);
    cur_params.push_back(var(type, n->id, true));
}

// Enter compound statement scope
void scopevisitor::visit(int d, compound_stmt_node* n)
{
    enter_scope();
    if (n->decls) n->decls->accept(d,*this);
    if (n->body) n->body->accept(d,*this);
    leave_scope();
}
void scopevisitor::visit(int d, loc_decl_node* n)
{
    if (n->head) n->head->accept(d,*this);
    n->tail->accept(d,*this);
}

void scopevisitor::visit(int d, stmt_list_node* n)
{
    if(n->head) n->head->accept(d, *this);
    if(n->tail) n->tail->accept(d, *this);
}

void scopevisitor::visit(int d, expr_stmt_node* n)
{
    if(n->expr) n->expr->accept(d+1, *this);
}

// Enter conditional statement scope
void scopevisitor::visit(int d, select_stmt_node* n)
{
    n->expr->accept(d,*this);
    n->ifstmt->accept(d+1,*this);
    if (n->elsestmt)
        n->elsestmt->accept(d+1,*this);
}

void scopevisitor::visit(int d, iter_stmt_node* n)
{
    n->expr->accept(d,*this);
    n->body->accept(d,*this);
}

// Return statement. This is _only_ valid with certain symbolic return values!
void scopevisitor::visit(int d, return_stmt_node* n)
{
    using namespace std;

    stringstream ss;
    ofstream of("/dev/null");
    codevisitor cv(ss, of);
    n->expr->accept(0, cv);

    string s = ss.str();
    if(s != "CONTINUE" and s != "KILLNODE" and s != "TERMINATE")
    {
        cout << n->line << ": return expression must be one of "
             << "CONTINUE, KILLNODE, and TERMINATE" << endl;
        m_good = false;
    }
}

// Check compatible types for assignment experssion
void scopevisitor::visit(int d, assign_expr_node* n)
{
    n->var->accept(d, *this);
    n->expr->accept(d, *this);

    if(!n->var->type){
        std::cout <<n->line <<": warning: type of lhs cannot be verified" <<std::endl;
    }else if(!n->expr->type){
        std::cout <<n->line <<": warning: type of rhs cannot be verified" <<std::endl;
    }else if(!check_types(n->var->type,n->expr->type))
    {
        std::cout << n->var->type << " " << n->expr->type << std::endl;
        std::cout << n->line << ": error: incompatible types!" << std::endl;
        m_good = false;
    }
    n->type = n->var->type;
}

// Check variable references. This checks types and warns of symbols that
// either are not declared at all, or are declared outside of (checked)
// RTSync scopes and can therefore not be type checked at all
void scopevisitor::visit(int d, simple_var_node* n)
{
    using namespace std;
    declarable* dec = find_symbol(n->id).get();
    if(!dec)
    {
        cout << n->line << ": warning: " << n->id << " not declared in an "
             << "RTSync scope" << endl;
        return;
    }

    var* v = dynamic_cast<var*>(dec);
    if(!v)
    {
        cout << n->line << ": error: " << n->id << " is a method: expected "
             << "a variable" << endl;
        m_good = false;
        return;
    }
    if(v->array)
    {
        cout << n->line << ": error: " << n->id << " is an array: expected "
             << "a simple variable" << endl;
        m_good = false;
        return;
    }

    if(v->type == "int")
        n->type = expr_node::EX_INT;
    else if(v->type == "float")
        n->type = expr_node::EX_FLOAT;
    else if(v->type == "string")
        n->type = expr_node::EX_STRING;
    else
        n->type = expr_node::EX_INV;
}

// Check variable references. This checks types and warns of symbols that
// either are not declared at all, or are declared outside of (checked)
// RTSync scopes and can therefore not be type checked at all
void scopevisitor::visit(int d, arr_var_node* n)
{
    using namespace std;
    declarable* dec = find_symbol(n->id).get();
    if(!dec)
    {
        cout << n->line << ": warning: " << n->id << " not declared in an "
             << "RTSync scope" << endl;
        return;
    }

    var* v = dynamic_cast<var*>(dec);
    if(!v)
    {
        cout << n->line << ": error: " << n->id << " is a method: expected "
             << "an array" << endl;
        m_good = false;
        return;
    }
    if(!v->array)
    {
        cout << n->line << ": error: " << n->id << " is an simple variable: "
             << "expected an array" << endl;
        m_good = false;
        return;
    }

    if(v->type == "int")
        n->type = expr_node::EX_INT | expr_node::EX_ARRAY;
    else if(v->type == "float")
        n->type = expr_node::EX_FLOAT | expr_node::EX_ARRAY;
    else if(v->type == "string")
        n->type = expr_node::EX_STRING | expr_node::EX_ARRAY;
    else
        n->type = expr_node::EX_INV;
}

// Check types in expressions
void scopevisitor::visit(int d, compound_simple_expr_node* n)
{
    n->expr1->accept(d,*this);
    n->rel->accept(d,*this);
    n->expr2->accept(d,*this);
    if(!n->expr1->type || !n->expr2->type){
        std::cout << n->line << ": warning: type of expression cannot be verified"
            <<std::endl;
    }else if (!(n->type = check_types(n->expr1->type,n->expr2->type)))
    {
        std::cout << n->line << ": error: incompatible types!" << std::endl;
        m_good = false;
    }
}

// Check types in expressions
void scopevisitor::visit(int d, compound_add_expr_node* n)
{
    n->expr1->accept(d,*this);
    n->op->accept(d,*this);
    n->expr2->accept(d,*this);
    if (!n->expr1->type || !n->expr2->type){
        std::cout << n->line << ": warning: type of expression cannot be verified"
           << std::endl;
    }else if (!(n->type = check_types(n->expr1->type,n->expr2->type)))
    {
        std::cout << n->line << ": error: incompatible types!" << std::endl;
        m_good = false;
    }
}

// Check types in expressions
void scopevisitor::visit(int d, compound_term_node* n)
{
    n->expr1->accept(d,*this);
    n->op->accept(d,*this);
    n->expr2->accept(d,*this);
    if (!(n->type = check_types(n->expr1->type,n->expr2->type)))
    {
        std::cout << n->line << ": error: incompatible types!" << std::endl;
        m_good = false;
    }
}

void scopevisitor::visit(int d, paren_factor_node* n)
{
    n->expr->accept(d,*this);
    n->type = n->expr->type;
}

void scopevisitor::visit(int d, var_factor_node* n)
{
    n->var->accept(d,*this);
    n->type = n->var->type;
}

// Check function call, arguments, etc.
void scopevisitor::visit(int d, call_node* n)
{
    n->type = expr_node::EX_UNKNOWN;

    using namespace std;
    declarable* dec = find_symbol(n->id1).get();
    if(!dec)
    {
        cout << n->line << ": warning: " << n->id1 << " not declared in an "
             << "RTSync scope" << endl;
        return;
    }

    func* f = dynamic_cast<func*>(dec);
    if(!f)
    {
        cout << n->line << ": error: " << n->id1 << " is a variable: expected "
             << "a function or method" << endl;
        m_good = false;
        return;
    }

    cur_args.clear();
    if(n->args)
        n->args->accept(d,*this);

    if(f->args.size() != cur_args.size())
    {
        std::cout << n->line << ": error: wrong number of arguments to"
                  << f->name << std::endl;
        m_good = false;
        return;
    }

    paramlist::iterator p_it = f->args.begin();
    std::list<int>::iterator a_it = cur_args.begin();
    while(a_it != cur_args.end())
    {
        if(!match(*a_it, p_it->type, p_it->array))
        {
            std::cout << n->line << ": error: parameter " << p_it->name
                      << " does not match parameter type" << std::endl;
            m_good = false;
        }
        ++a_it;
        ++p_it;
    }
}

void scopevisitor::visit(int d, args_node* n)
{
    n->args->accept(d,*this);
}

// Add argument to current arg list
void scopevisitor::visit(int d, arg_list_node* n)
{
    if (n->head) n->head->accept(d,*this);
    n->tail->accept(d,*this);

    cur_args.push_back(n->tail->type);
}

void scopevisitor::visit(int d, actor_node* n)
{
    enter_scope();
    n->tail->accept(d+1,*this);
    leave_scope();
}

void scopevisitor::visit(int d, global_sync_node* n)
{
    enter_scope();
    n->tail->accept(d+1,*this);
    leave_scope();
}

void scopevisitor::visit(int d, sync_node* n)
{
    enter_scope();
    n->tail->accept(d+1,*this);
    leave_scope();
}

void scopevisitor::visit(int d, init_node* n)
{
    n->head->accept(d+1,*this);
}

void scopevisitor::visit(int d, sync_init_node* n)
{
    n->head->accept(d,*this);
    n->tail->accept(d,*this);
}

void scopevisitor::visit(int d, sync_decl_list_node* n)
{
    if(n->vars)
        n->vars->accept(d+1,*this);
    n->init->accept(d+1,*this);
    n->constraints->accept(d+1,*this);
    n->triggers->accept(d+1,*this);
    n->exception->accept(d+1,*this);
}

void scopevisitor::visit(int d, constraint_block_node* n)
{
    if(n->head) n->head->accept(d+1,*this);
}

void scopevisitor::visit(int d, constraint_list_node* n)
{
    if (n->head) n->head->accept(d,*this);
    n->tail->accept(d,*this);
}

void scopevisitor::visit(int d, constraint_node* n)
{
    n->tail->accept(d,*this);
}

void scopevisitor::visit(int d, trigger_block_node* n)
{
    if (n->head) n->head->accept(d+1,*this);
}

void scopevisitor::visit(int d, trigger_list_node* n)
{
    if (n->head) n->head->accept(d,*this);
    n->tail->accept(d,*this);
}

void scopevisitor::visit(int d, ed_trigger_node* n)
{
    n->expr->accept(d,*this);
}

void scopevisitor::visit(int d, action_trigger_node* n)
{
    n->tail->accept(d,*this);
}

void scopevisitor::visit(int d, actor_params_node* n)
{
    if (n->head) n->head->accept(d,*this);
    n->tail->accept(d,*this);
}

void scopevisitor::visit(int d, var_decls_node* n)
{
    if (n->head) n->head->accept(d,*this);
    n->tail->accept(d,*this);
}

void scopevisitor::visit(int d, func_decls_node* n)
{
    if (n->head) n->head->accept(d,*this);
    n->tail->accept(d,*this);
}

void scopevisitor::visit(int d, module_list_node* n)
{
    if (n->head) n->head->accept(d,*this);
    n->tail->accept(d,*this);
}

void scopevisitor::visit(int d, call_stmt_node* n)
{
    if(n->args)
        n->args->accept(d,*this);
}

void scopevisitor::visit(int d, sync_call_stmt_node* n)
{
    if(n->args)
        n->args->accept(d,*this);
}
