/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef SCOPEVISITOR_H
#define SCOPEVISITOR_H

#include "visitor.h"

#include <list>
#include <map>
#include <string>
#include <boost/shared_ptr.hpp>


// The scope visitor is used to perform scope and type checking on an AST.
class scopevisitor : public visitor
{
    // Final status
    bool m_good;

    // Declarable entity: Base class for variable and function declarations
    struct declarable
    {
        std::string name;
        declarable(const std::string& n) : name(n) { }
        virtual ~declarable() { }
    };

    // Variable
    struct var : public declarable
    {
        std::string type;
        bool array;

        var(const std::string& t, const std::string& n, bool a=false)
        : declarable(n), type(t), array(a) { }
    };

    // Parameters of the function currently being processed
    typedef std::list<var> paramlist;
    paramlist cur_params;
    // Arguments being parsed whilst parsing a function call
    std::list<int> cur_args;

    // Function declaration
    struct func : public declarable
    {
        paramlist args;

        func(const std::string& n, const paramlist& a)
        : declarable(n), args(a) { }
    };

    // Map symbol names to declarable entities
    typedef std::map<std::string, boost::shared_ptr<declarable> > table;
    std::list<table> symbol_table;

    boost::shared_ptr<declarable> find_symbol(const std::string& name);

    // Functionality to perform when entering and leaving scopes. Note that
    // while presently nearly no-ops, they have intermittently had more
    // onerous duties.

    void enter_scope()
    {
        symbol_table.push_back(table());
    }

    void leave_scope()
    {
        symbol_table.pop_back();
    }

public:
    scopevisitor() : m_good(true) { }
    ~scopevisitor() { }

    bool good() const
    {
        return m_good;
    }

    // Check variable declaration against multiple declarations
    void visit(int d, simple_var_decl_node* n);
    void visit(int d, arr_var_decl_node* n);

    // Enter a function definition: Set up scopes and symbol tables
    void visit(int d, fun_decl_node* n);

    // Enter compound statement scope
    void visit(int d, compound_stmt_node* n);

    // Enter conditional statement scope
    void visit(int d, select_stmt_node* n);

    // Return statement. This is _only_ valid with certain symbolic return
    // values!
    void visit(int d, return_stmt_node* n);

    // Check compatible types for assignment experssion
    void visit(int d, assign_expr_node* n);

    // Check variable references. This checks types and warns of symbols that
    // either are not declared at all, or are declared outside of (checked)
    // RTSync scopes and can therefore not be type checked at all
    void visit(int d, simple_var_node* n);
    void visit(int d, arr_var_node* n);

    // Check types in expressions
    void visit(int d, compound_simple_expr_node* n);
    void visit(int d, compound_add_expr_node* n);
    void visit(int d, compound_term_node* n);

    // Check function call, arguments, etc.
    void visit(int d, call_node* n);

    // Add argument to current arg list
    void visit(int d, arg_list_node* n);

    // Enter appropriate scopes
    void visit(int d, actor_node* n);
    void visit(int d, global_sync_node* n);
    void visit(int d, sync_node* n);

    // The scope visitor's task is very limited, and as such its visits to
    // most node types consists simply of visiting the node's children. Such
    // code will not be commented upon. Expression nodes will propagate their
    // types upward: This is also omitted.

    void visit(int d, program_node* n);
    void visit(int d, decl_list_node* n);
    void visit(int d, exception_block_node* n);
    void visit(int d, param_list_node* n);
    void visit(int d, simple_param_node* n);
    void visit(int d, arr_param_node* n);
    void visit(int d, loc_decl_node* n);
    void visit(int d, stmt_list_node* n);
    void visit(int d, expr_stmt_node* n);
    void visit(int d, iter_stmt_node* n);
    void visit(int d, paren_factor_node* n);
    void visit(int d, var_factor_node* n);
    void visit(int d, args_node* n);

    void visit(int d, init_node* n);
    void visit(int d, sync_init_node* n);
    void visit(int d, sync_decl_list_node* n);
    void visit(int d, constraint_block_node* n);
    void visit(int d, constraint_list_node* n);
    void visit(int d, constraint_node* n);
    void visit(int d, trigger_block_node* n);
    void visit(int d, trigger_list_node* n);
    void visit(int d, ed_trigger_node* n);
    void visit(int d, action_trigger_node* n);
    void visit(int d, actor_params_node* n);
    void visit(int d, var_decls_node* n);
    void visit(int d, func_decls_node* n);
    void visit(int d, module_list_node* n);
    void visit(int d, call_stmt_node* n);
    void visit(int d, sync_call_stmt_node* n);

    // The functions below are all no-ops and are merely included and
    // implemented to conform to the visitor interface. They will not be
    // called.

    void visit(int d, node* n) { }
    void visit(int d, decl_node* n) { }
    void visit(int d, var_decl_node* n) { }
    void visit(int d, type_spec_node* n) { }
    void visit(int d, params_node* n) { }
    void visit(int d, param_node* n) { }
    void visit(int d, stmt_node* n) { }
    void visit(int d, expr_node* n) { }
    void visit(int d, var_node* n) { }
    void visit(int d, simple_expr_node* n) { }
    void visit(int d, relop_node* n) { }
    void visit(int d, add_expr_node* n) { }
    void visit(int d, addop_node* n) { }
    void visit(int d, term_node* n) { }
    void visit(int d, mulop_node* n) { }
    void visit(int d, factor_node* n) { }
    void visit(int d, num_factor_node* n) { }
    void visit(int d, trigger_node* n) { }
    void visit(int d, actor_param_node* n) { }
    void visit(int d, module_node* n) { }
    void visit(int d, pass_node* n) { }
};

#endif // #define SCOPEVISITOR_H
