/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef VISITOR_H
#define VISITOR_H

#include "ast_decls.h"

#include <string>

// For code indentation (I'll grant you that it's fairly useless, but
// debugging code generation is much easier when the generated code is
// readable)
inline std::string indent(const int n)
{
    return std::string(n*2, ' ');
}

// Abstract base class for visitors of the AST (see ast.h/cc); used to
// implement the Visitor Pattern. If you are not familiar with the visitor
// pattern, it is recommended that you familiarize yourself with it before
// examining this, let alone implementing visitors. The canonical resource is
// 'Design Patterns -- Elements of Reusable Object-Oriented Software' by
// Erich Gamma, Richard Helm, Ralph Johnson, and John Vlissides (commonly
// known as "the Gang of Four book"), ISBN 0-201-63361-2.
class visitor
{
public:
    virtual ~visitor() { }

    virtual void visit(int d, node* n) = 0;
    virtual void visit(int d, program_node* n) = 0;
    virtual void visit(int d, decl_list_node* n) = 0;
    virtual void visit(int d, decl_node* n) = 0;
    virtual void visit(int d, var_decl_node* n) = 0;
    virtual void visit(int d, simple_var_decl_node* n) = 0;
    virtual void visit(int d, arr_var_decl_node* n) = 0;
    virtual void visit(int d, type_spec_node* n) = 0;
    virtual void visit(int d, exception_block_node* n) = 0;
    virtual void visit(int d, fun_decl_node* n) = 0;
    virtual void visit(int d, params_node* n) = 0;
    virtual void visit(int d, param_list_node* n) = 0;
    virtual void visit(int d, param_node* n) = 0;
    virtual void visit(int d, simple_param_node* n) = 0;
    virtual void visit(int d, arr_param_node* n) = 0;
    virtual void visit(int d, compound_stmt_node* n) = 0;
    virtual void visit(int d, loc_decl_node* n) = 0;
    virtual void visit(int d, stmt_list_node* n) = 0;
    virtual void visit(int d, stmt_node* n) = 0;
    virtual void visit(int d, expr_stmt_node* n) = 0;
    virtual void visit(int d, select_stmt_node* n) = 0;
    virtual void visit(int d, iter_stmt_node* n) = 0;
    virtual void visit(int d, return_stmt_node* n) = 0;
    virtual void visit(int d, expr_node* n) = 0;
    virtual void visit(int d, assign_expr_node* n) = 0;
    virtual void visit(int d, var_node* n) = 0;
    virtual void visit(int d, simple_var_node* n) = 0;
    virtual void visit(int d, arr_var_node* n) = 0;
    virtual void visit(int d, simple_expr_node* n) = 0;
    virtual void visit(int d, compound_simple_expr_node* n) = 0;
    virtual void visit(int d, relop_node* n) = 0;
    virtual void visit(int d, add_expr_node* n) = 0;
    virtual void visit(int d, compound_add_expr_node* n) = 0;
    virtual void visit(int d, addop_node* n) = 0;
    virtual void visit(int d, term_node* n) = 0;
    virtual void visit(int d, compound_term_node* n) = 0;
    virtual void visit(int d, mulop_node* n) = 0;
    virtual void visit(int d, factor_node* n) = 0;
    virtual void visit(int d, paren_factor_node* n) = 0;
    virtual void visit(int d, var_factor_node* n) = 0;
    virtual void visit(int d, num_factor_node* n) = 0;
    virtual void visit(int d, call_node* n) = 0;
    virtual void visit(int d, args_node* n) = 0;
    virtual void visit(int d, arg_list_node* n) = 0;

    virtual void visit(int d, actor_node* n) = 0;
    virtual void visit(int d, global_sync_node* n) = 0;
    virtual void visit(int d, sync_node* n) = 0;
    virtual void visit(int d, init_node* n) = 0;
    virtual void visit(int d, sync_init_node* n) = 0;
    virtual void visit(int d, sync_decl_list_node* n) = 0;
    virtual void visit(int d, constraint_block_node* n) = 0;
    virtual void visit(int d, constraint_list_node* n) = 0;
    virtual void visit(int d, constraint_node* n) = 0;
    virtual void visit(int d, trigger_block_node* n) = 0;
    virtual void visit(int d, trigger_list_node* n) = 0;
    virtual void visit(int d, trigger_node* n) = 0;
    virtual void visit(int d, ed_trigger_node* n) = 0;
    virtual void visit(int d, action_trigger_node* n) = 0;
    virtual void visit(int d, actor_params_node* n) = 0;
    virtual void visit(int d, actor_param_node* n) = 0;
    virtual void visit(int d, var_decls_node* n) = 0;
    virtual void visit(int d, func_decls_node* n) = 0;
    virtual void visit(int d, module_list_node* n) = 0;
    virtual void visit(int d, module_node* n) = 0;
    virtual void visit(int d, pass_node* n) = 0;

    virtual void visit(int d, call_stmt_node* n) = 0;
    virtual void visit(int d, sync_call_stmt_node* n) = 0;
};

#endif // #define VISITOR_H
