/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "actor.h"
#include "postmaster.h"
#include "logger.h"
#include "messages.h"

#include <algorithm>


actor::actor(const actorref& n, postmaster& p)
: rt_current(rt_inf), rt_highest(rt_inf), name(n), PostMaster(p)
{
    // Use default attributes, else we need to setup a pthread_mutexattr_t
    pthread_mutex_init(&m_queue_mutex, 0);
}


// Method for synchronizers to notify the actor that a message of the
// specified priority is waiting at synchronizer s
void actor::register_msg(syncref s, rttime_t priority)
{
    rt_log << s << " notified " << name << " a message is available\n";
    synchstruct s1 = { s, priority };

    pthread_mutex_lock(&m_queue_mutex);
    m_msg_requests.enqueue(s1);
    if(m_msg_requests.size())
        rt_highest = m_msg_requests.peek().priority;
    else
        rt_highest = rt_inf;
    pthread_mutex_unlock(&m_queue_mutex);
}

// Notify all listeners that an event occurred
void actor::notify(const method& m, const rttime_t& t)
{
    for(std::vector<syncref>::iterator it = listeners.begin();
        it != listeners.end(); ++it)
    {
        rt_log << name << " notifying " << *it << " that it "
                  << "completed " << m << " at " << t << "\n";
        PostMaster.sm_notify_action(*it, name, m, t);
    }
}

// Method for synchronizers to register themselves as listeners
void actor::register_listener(const syncref& s)
{
    rt_log << "Actor " << name << ": " << s << " is listening" << "\n";
    if(std::find(listeners.begin(), listeners.end(), s) == listeners.end())
        listeners.push_back(s);
}

// Returns the actor's present scheduling priority in the form of a
// deadline; the lower, the more urgent.
rttime_t actor::getpriority() const
{
    // If we have a more urgent current priority, use it
    return std::min(rt_highest, rt_current);
}

// Polling functionality below

// To pass to pthread_create
void* start_polling_func_a(actor* a)
{
    rt_log << "(actor::)start_polling_func()" << "\n";
    a->poll();
    return 0;
}

// Start the polling thread; each derived constructor shall call this once
// its members are initialized
void actor::start_polling()
{
    rt_log << "actor::start_polling()" << "\n";
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

    int result = pthread_create(&m_pollthread, &attr,
                                (void* (*)(void*)) start_polling_func_a,
                                (void*) this);
#ifdef USE_SCHED
    sched_register(m_pollthread);
#endif
    rt_log << name << " is " << m_pollthread << "\n";

    if(result)
    {
        rt_log << "Error: Actor failed to create polling thread" << "\n";
        throw std::string("Error: Actor failed to create polling thread");
    }

    rt_log << "result: " << result << "\n";
    rt_log << "actor::start_polling() completed" << "\n";
}

// Kill the polling thread; each derived destructor shall call this before
// destroying any of its members
void actor::end_polling()
{
    pthread_kill(m_pollthread, 9);
}

// Master polling function; iterates over the messages in the actor's
// internal queue and dispatches them in priority order.
void actor::poll()
{
    rt_log << name << " starts its polling ..." << "\n";
    while(true)
    {
        pthread_mutex_lock(&m_queue_mutex);
        if(m_msg_requests.size())
        {
            // Fetch highest-priority message
            synchstruct ss = m_msg_requests.dequeue();
            if(m_msg_requests.size())
                rt_highest = m_msg_requests.peek().priority;
            else
                rt_highest = rt_inf;
            pthread_mutex_unlock(&m_queue_mutex);

            // Get the message
            message m = PostMaster.sm_request_msg(ss.s, name);

            // The message may be null, if the method has been flagged as
            // disabled. If so, put the request back in the queue; else
            // process the message.
            if(m != null_msg)
            {
                m.timestamp.received = rt_gettime();
                rt_log << m;
                execute(m);
            }
            else
            {
                pthread_mutex_lock(&m_queue_mutex);
                m_msg_requests.enqueue(ss);
                if(m_msg_requests.size())
                    rt_highest = m_msg_requests.peek().priority;
                else
                    rt_highest = rt_inf;
                pthread_mutex_unlock(&m_queue_mutex);
                sched_yield();
            }
        }
        // If we have nothing to do, yield the scheduler slice.
        else
        {
            pthread_mutex_unlock(&m_queue_mutex);
            sched_yield();
        }
    }
}
