/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef ACTOR_H
#define ACTOR_H

#include <pthread.h>
#include <vector>

#include "timing.h"
#include "pqueue.h"
#include "schedulable.h"
#include "reftypes.h"

class postmaster;
class message;
class actor;


// Simple struct to track message requests by priority
struct synchstruct
{
    syncref s;
    rttime_t priority;

    bool operator<(const synchstruct& s) {
        return priority < s.priority;
    }
};



// This provides the abstract base class for actors (see docs/actor_concepts).
// Concrete actors are generated from user RTSync code and shall provide
// suitable constructors and an implementation for the virtual execute()
// function.
//   An actor basically functions by associating itself with a postmaster and
// setting up a polling thread. External entities may, through the postmaster,
// feed it new messages; the polling thread shall iterate and select suitable
// messages from an internal queue, which the execute() function dispatches.
class actor : public rt_sched_actor
{
    // Thread id of polling thread
    pthread_t m_pollthread;

    // Master polling function; iterates over the messages in the actor's
    // internal queue and dispatches them in priority order.
    void poll();

    // Message requests (external entities do not queue messages, but
    // notifications that messages are waiting)
    pthread_mutex_t m_queue_mutex;
    pqueue<synchstruct> m_msg_requests;
protected:
    rttime_t rt_current; // Priority of message being processed
    rttime_t rt_highest; // Highest priority in queue

    const actorref name;
    postmaster& PostMaster;

    // Keep track of listeners (for action triggers etc.)
    std::vector<syncref> listeners;

    // Notify all listeners that an event occurred
    void notify(const method& m, const rttime_t& t);

    // Start the polling thread; each derived constructor shall call this once
    // its members are initialized
    void start_polling();

    // Kill the polling thread; each derived destructor shall call this before
    // destroying any of its members
    void end_polling();

    // Hook to pass to pthread_create; can't use something with thiscall
    // calling conventions
    friend void* start_polling_func_a(actor* a);

    // Process message - inheriting classes must pass it on to its methods,
    // as appropriate. This function shall
    //   - Update the actor's current priority to reflect the running message
    //   - Execute the proper internal method using the parameters encoded in
    //     the message
    //   - Notify the external entities of message execution by calling
    //     PostMaster.sm_ack_msg() with suitable timestamps
    //   - If the message is flagged as being synchronous, notify the
    //     postmaster of completion by calling PostMaster.recvsyncmsg() to
    //     unblock the caller
    //   - Notify all listeners using the notify() function
    //   - Reset the actor's current priority
    virtual void execute(message& m) = 0;

    friend class postmaster;
public:
    actor(const actorref& name, postmaster& postmaster);
    virtual ~actor() { }

    // Returns the actor's present scheduling priority in the form of a
    // deadline; the lower, the more urgent.
    rttime_t getpriority() const;

    // Method for synchronizers to notify the actor that a message of the
    // specified priority is waiting at synchronizer s
    void register_msg(syncref s, rttime_t priority);

    // Method for synchronizers to register themselves as listeners
    void register_listener(const syncref& s);
};


#endif // #define ACTOR_H
