/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef GLOBALSYNC_H
#define GLOBALSYNC_H

#include <set>
#include <memory>
#include <pthread.h>

#include "actionlistener.h"
#include "schedulable.h"


class postmaster;


// Abstract base class for "global" synchronizers. While conceptually global
// in scope, and historically global entities, each node in a distributed
// system may possess a global synchronizer. What distinguishes a global
// synchronizer from a normal one is that
//   - the global synchronizer does not act as a co-ordinating entity
//     (see docs/sync_concepts)
//   - the global synchronizer checks and responds to constraints that are not
//     necessarily point-to-point, but may have intermediate steps.
// Concrete global synchronizers are generated from user code and shall
// provide suitable constructors and implementations for poll_triggers() and
// constraint_check().
//   A global synchronizer operates by associating itself with a postmaster
// and registering itself as a listener to all pertinent actors.
class globalsync : public actionlistener, public rt_sched_sync
{
    // Thread id for polling thread
    pthread_t m_pollthread;

public:
    // Internal event notification
    struct notification
    {
        actorref a;
        method m;
        rttime_t time;
    };

    // Master polling function; iterates over internal data structures.
    void poll_master();

    // Utility function: Wraps member function to start with pthread_create
    friend void* start_polling_func_gs(globalsync* s);

protected:

    postmaster& PostMaster;

    typedef std::set<notification> notification_list;
    // Notifications of min and max constraints
    notification_list min_notes, max_notes;
    pthread_mutex_t m_notify_mutex;

    // The right-hand side (RHS) of a constraint is an event that has happened
    // in the system.
    notification_list rhs;
    pthread_mutex_t m_rhs_mutex;

    syncref name;

    // Start the polling thread; each derived constructor shall call this once
    // its members are initialized
    void start_polling();
    // Kill the polling thread; each derived destructor shall call this before
    // destroying any of its members
    void end_polling();

    // Poll enable/disable triggers; do not call this explicitly. This
    // function shall examine certain conditions and set and propagate the
    // enabled/disabled status of actor methods appropriately.
    //   This shall poll each trigger only once; it will be called in a loop.
    virtual void poll_triggers() = 0;

    // Exception handler used to process timing exceptions thrown by
    // poll_msg(). This can execute arbitrary code, but shall return one of
    // the symbolic constants below.
    virtual int exception_handler(const timing_exception&) = 0;

    // Symbolic constants for exception responses
    static const int TERMINATE = 99;     // Terminate the system
    static const int KILLNODE  = 443556; // Terminate the local node
    static const int CONTINUE  = 0;      // Proceed as before

public:
    globalsync(const syncref& ref, postmaster& pm);
    virtual ~globalsync() { }

    // This is the globalsync equivalent of the synchronizer's message
    // polling. The difference is that while a point-to-point synchronizer
    // enforces constraints, the global synchronizer merely analyses events
    // after the fact and determines whether any constraints were violated.
    //   This function shall iterate over all the _right hand sides_ of all
    // constraints, and for each such right hand side, ensure that the rhs
    // (which corresponds to an event that has happened in the system) did
    // not violate any timing constraints, as defined by min and max notes.
    virtual void constraint_check() = 0;
};

inline bool operator==(const globalsync::notification& n1,
                       const globalsync::notification& n2)
{
    return n1.a == n2.a and n1.m == n2.m;
}

inline bool operator<(const globalsync::notification& n1,
                      const globalsync::notification& n2)
{
    return n1.a != n2.a? n1.a < n2.a : n1.m < n2.m;
}


#endif // #define GLOBALSYNC_H
