/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef LOCK_LIST_H
#define LOCK_LIST_H

#include <list>
#include "mutex.h"

// List which inherits from mutex to provide synchronization capabilities
//
// Guarantees:
// - Inserting or erasing elements does not invalidate any iterators (except,
//   in the case of deletion, iterators to the element being erased).
template <class T>
class lock_list : public mutex
{
    typedef std::list<T> list_type;
    list_type m_list;
public:
    typedef typename list_type::iterator       iterator;
    typedef typename list_type::const_iterator const_iterator;
    typedef typename list_type::size_type      size_type;
    typedef T value_type;

    iterator begin() { return m_list.begin(); }
    iterator end()   { return m_list.end(); }
    const_iterator begin() const { return m_list.begin(); }
    const_iterator end() const   { return m_list.end(); }

    void push_front(const T& t) { m_list.push_front(t); }
    void push_back(const T& t) { m_list.push_back(t); }
    void pop_front() { m_list.pop_front(); }
    void pop_back() { m_list.pop_back(); }

    size_type size() const { return m_list.size(); }
    void clear() { m_list.clear(); }

    size_type erase(const value_type& k) {
        size_type t = size();
        m_list.remove(k);
        return t - size();
    }
    iterator erase(iterator pos) { return m_list.erase(pos); }

    iterator find(const value_type& k) {
        iterator it = m_list.begin();
        while(it != m_list.end()){
            if(*it == k) return it;
            it++;
        }
        return it;
    }
};


#endif // #define LOCK_LIST_H
