/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef LOCK_MAP_H
#define LOCK_MAP_H

#include <map>
#include "mutex.h"

// Map which inherits from mutex to provide synchronization capabilities
//
// Guarantees:
// - Erasing an element does not invalidate iterators other than to the
//   element being erased
// - Inserting an element does not invalidate any iterators
//
// In the future the std::map<> (usually implemented as a red-black tree)
// could be replaced by e.g. a hash table, provided that the above guarantees
// are maintained.
template <class S, class T>
class lock_map : public mutex
{
    typedef std::map<S, T> map_type;
    std::map<S, T> m_map;
public:
    typedef typename map_type::iterator       iterator;
    typedef typename map_type::const_iterator const_iterator;
    typedef typename map_type::size_type      size_type;
    typedef S key_type;
    typedef T data_type;

    iterator begin() { return m_map.begin(); }
    iterator end()   { return m_map.end(); }
    const_iterator begin() const { return m_map.begin(); }
    const_iterator end() const   { return m_map.end(); }

    size_type size() const { return m_map.size(); }

    size_type erase(const key_type& k) { return m_map.erase(k); }
    void erase(iterator pos) { m_map.erase(pos); }

    iterator find(const key_type& k) { return m_map.find(k); }
    data_type& operator[](const key_type& k) { return m_map[k]; }
};


#endif // #define LOCK_MAP_H
