/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef LOGGER_H
#define LOGGER_H

#include <iostream>
#include <fstream>
#include <pthread.h>
#include "messages.h"

/*
 * Logger class for debug output.  Output is not mutex protected, so
 * occassional interleaving may be noticed.  This decision was made to ensure
 * the most efficiency was maintained.
 *
 * Interleaving is extremely minimal in practice.
 *
 * In all levels except 0, a file called messages.log is created which holds
 * output of all messages from the previous execution.
 *
 * The fastlog option writes without explicity flushing the stream.
 *
 * rt_log exists upon execution.
 */
extern class rt_logger
{
    std::ofstream ofile;
    std::ostream& os;

    // Log levels:
    //  0 - log nothing
    //  1 - log only to file
    //  2 - log to file, print only messages
    //  3 - log to file, print everything
    int loglevel;
    bool fastlog;

public:
    rt_logger(std::ostream& ostream, int level=1, bool fast=false)
    : ofile("messages.log"), os(ostream), loglevel(level), fastlog(fast)
    { }

    // Sets the log level - see above for a description of levels.
    void setlevel(int level)
    {
        loglevel = level;
    }

    // Output for standard types
    template <typename T> rt_logger& operator<<(const T& t)
    {
        if(loglevel >= 3)
        {
            os << t;
            if(!fastlog)
                os << std::flush;
        }
        return *this;
    }

    // Output for stream manipulators like endl
    rt_logger& operator<<(std::ostream& (*manip)(std::ostream&))
    {
        if(loglevel >= 3)
        {
            os << manip;
            if(!fastlog)
                os << std::flush;
        }
        return *this;
    }

    // Output for a message
    rt_logger& operator<<(const message& m)
    {
        if(loglevel >= 1)
        {
            ofile << m << "\n";
            if(!fastlog)
                ofile << std::flush;
        }
        if(loglevel >= 2)
        {
            os << m << "\n";
            if(!fastlog)
                os << std::flush;
        }
        return *this;
    }

} rt_log;


#endif // #define LOGGER_H
