/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef MESSAGES_H
#define MESSAGES_H

#include <stdexcept>
#include <string>
#include <sstream>
#include <pthread.h>

#include "net_id.h"
#include "mutex.h"
#include "timing.h"
#include "reftypes.h"
#include "params.h"


// Message timestamp structure
struct timestamp_t
{
    // 1. Created by actor (VITAL)
    rttime_t creation;
    // 2. Received by synchronizer, queued for transmission (DEBUG)
    rttime_t dispatch;
    // 3. Received by receiver actor (VITAL)
    rttime_t received;
    // 4. Execution/processing finished (DEBUG)
    rttime_t executed;
};


// This is the message used in all communication between actors. It encodes
// a volume of critical information, such as sender actor and method, receiver
// actor and method, synchronizing entity, and various timestamps in the life
// of the message.
class message
{
public:
    // A unique serial number for a message. In order to guarantee uniqueness
    // across a networked system, the local net_id is part of the serial
    // number.
    class serial_no
    {
        net_id nid;
        unsigned long seq;
    public:
        serial_no(const net_id& n, unsigned long s) : nid(n), seq(s) { }

        bool operator==(const serial_no& s) const {
            return nid == s.nid and seq == s.seq;
        }

        bool operator!=(const serial_no& s) const {
            return !(*this == s);
        }

        bool operator<(const serial_no& s) const {
            return nid < s.nid or (nid == s.nid and seq < s.seq);
        }

        friend std::istream& operator>>(std::istream&, serial_no&);
        friend std::ostream& operator<<(std::ostream&, const serial_no&);
    };

    serial_no serial;

private:
    // Simple class for generating unique serial numbers for messages. The
    // serial numbers are unique and sequential; note, however, that messages
    // observed in communication may not be sequential since temporaries also
    // have unique serial numbers.
    class serial_generator
    {
        mutex mut;
        net_id nid;
        unsigned long seq;
    public:
        serial_generator() : seq(1) { }

        serial_no next_serial();
        // Specify net_id of the local system
        void set_nid(net_id n);
    };

    // We obviously need a static generator shared by all messages
    static serial_generator generator;
public:
    message() : serial(generator.next_serial()) { }

    static message null_msg() {
        message m;
        m.serial = serial_no(net_id(), 0);
        return m;
    }

    // Set generator's net_id
    static void set_nid(const net_id& n) {
        generator.set_nid(n);
    }

    // Deadline timestamps
    rttime_t must_run_before;
    rttime_t must_run_after;

    // Entities involved in transaction
    actorref sender;
    method caller;

    actorref receiver;
    method callee;

    syncref sync;

    // Method call parameters
    params m_params;

    // Miscellaneous timestamps
    timestamp_t timestamp;

    bool synchronous;

    message(actorref a1, method m1, actorref a2, method m2, params p)
    : serial(generator.next_serial()),
      must_run_before(rt_inf), must_run_after(rt_zero),
      sender(a1), caller(m1), receiver(a2), callee(m2),
      sync("null"), m_params(p), synchronous(false)
    {
        timestamp.creation = rt_gettime();
        timestamp.dispatch = timestamp.received = timestamp.executed = rt_zero;
    }

    bool operator==(const message& m) const {
        return serial == m.serial;
    }

    bool operator!=(const message& m) const {
        return serial != m.serial;
    }

    bool operator<(const message& m) const {
        return must_run_before < m.must_run_before;
    }
};

extern message null_msg;


// The stream operators are used to serialize messages for communication;
// the resultant strings are plain text but, while human readable, not
// necessarily human friendly -- they are designed to be easily parsed,
// not easily understood.

std::ostream& operator<<(std::ostream& os, const message& m);
std::istream& operator>>(std::istream& is, message& m);

std::ostream& operator<<(std::ostream& os, const message::serial_no& s);
std::istream& operator>>(std::istream& is, message::serial_no& s);


struct timing_exception
{
    message msg;

    timing_exception(const message& m) : msg(m)
    { }
};

std::ostream& operator<<(std::ostream& os, const timing_exception& te);


#endif // #define MESSAGES_H
