/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef NULLSYNC_H
#define NULLSYNC_H

#include "logger.h"
#include "messages.h"
#include "postmaster.h"
#include "timing.h"
#include "reftypes.h"


// Single implementation of the null synchronizer. This synchronizer is used
// whenever a synchronizer is needed between two actors, but no co-ordinating
// entity has been registered. It imposes no timing constrainst and does not
// hold messages, but merely passes them on.
class sync_null_t : public synchronizer
{
protected:
    // The null synchronizer does not enforce timing constraints and so does
    // nothing in the hypothetical event of a timing exception (but none will
    // be thrown).
    int exception_handler(const timing_exception& te) {
        return CONTINUE;
    }

public:

    // Null synchronizer is for unsynchronized messages, therefore the
    // immediate action is to remove it from the incoming queue and put it in
    // the incoming & notify destination actor about it.
    poll_result poll_msg(message& msg) {
        return REMOVE;
    }

    // The null synchronizer does not listen to anything
    void notify_action(const actorref& a, const method& m, const rttime_t& t) {
        ;
    }

    // The null synchronizer has no triggers
    void poll_triggers() {
        ;
    }

    sync_null_t(postmaster& p, const syncref& n) : synchronizer(n, p) {
        PostMaster.reg(n, this);
        start_polling();
    }

    ~sync_null_t() {
        rt_log << "Killing null synchronizer\n";
        end_polling();
    }
};


#endif // #define NULLSYNC_H
