/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef PARAMS_H
#define PARAMS_H

#include <iostream>
#include <sstream>
#include <vector>


// Parameter object for parameter passing through messages. A generic
// structure that acts like a stream (and uses a stream) to read from and
// write to generic types.
//   For every data type that a user wishes to use as an argument to an
// RTSync method, suitable stream operators must be defined to serialize and
// deserialize the object. Note that parameters therefore cannot be pointers,
// since the address would be encoded rather than any value; there is no
// guarantee that the recipient even resides on the same physical computer.
class params
{
    std::stringstream m_stream;
public:
    // Distinguished end marker
    static const std::string END;

    params() { }
    params(const params& p) : m_stream(p.m_stream.str()) { }

    params& operator=(const params& p) {
        m_stream.str(p.m_stream.str());
        return *this;
    }

    template <typename T>
    params& operator<<(const T& t) {
        m_stream << t;
        return *this;
    }

    template <typename T>
    params& operator>>(T& t) {
        m_stream >> t;
        return *this;
    }

    // The RTSync language uses std::vector for its implementation of arrays
    // (simpler given things like known length).
    template <typename T>
    params& operator<<(const std::vector<T>& v);

    template <typename T>
    params& operator>>(std::vector<T>& v);

    // Strings must be handled separately to encode parameter escape
    // characters
    params& operator<<(const std::string& s);

    params& operator<<(const char* c_str) {
        return this->operator<<(std::string(c_str));
    }

    params& operator>>(std::string& s);

    friend std::ostream& operator<<(std::ostream& os, const params& p);
    friend std::istream& operator>>(std::istream& is, params& p);
};


template <typename T>
params& params::operator<<(const std::vector<T>& v)
{
    for(typename std::vector<T>::const_iterator it = v.begin();
        it != v.end(); ++it)
    {
        m_stream << *it << " ";
    }
    m_stream << END << " ";
    return *this;
}

template <typename T>
params& params::operator>>(std::vector<T>& v)
{
    //m_stream << "vector ";
    std::string buf;
    while((m_stream >> buf))
    {
        if(buf == "$")
            break;
        else
        {
            std::stringstream ss(buf);
            T t;
            ss >> t;
            v.push_back(t);
        }
    }
    return *this;
}


#endif // #define PARAMS_H
