/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "postmaster.h"

#include "logger.h"

#include "actor.h"

#include "actionlistener.h"
#include "synchronizer.h"
#include "globalsync.h"
#include "nullsync.h"

#include "scheduler.h"
#include "tcp-utils.h"
#include "rtsocket.h"

#include <algorithm>
#include <sys/poll.h>

// Non member function to start the polling (called by new thread)
// Polling w.r.t. the postmaster is the section of code that checks all of
// the sockets and receives data on them.
void* start_polling_func(postmaster* p)
{
    p->net_recv();
    return 0;
}

/*
 * Functions of the Constructor:
 *   set up server socket
 *   create a net_id for ourselves
 *   let the message generator know that n_id
 *   name and set up a null synchronizer for this node
 *   connect to another postmaster (if specified)
 *     get id's of other postmasters
 *     set up net_ids for each of them
 *   start polling thread
 */
postmaster::postmaster(const char* host, unsigned short host_port,
                       unsigned short serverport, int backlog)
{
    using namespace std;
    using namespace boost;

    //disable messaging to screen
    rt_log.setlevel(1);

    // Create a server socket for this postmaster
    server_socket = passivesocket(serverport, backlog);
    if (server_socket < 0)
    {
        stringstream err;
        err << "Server socket creation error: " << server_socket << "\n";
        throw err.str().c_str();
    }
    else
        rt_log << "Created server socket on port: " << serverport << "\n";
    const char* localhost = getenv("HOSTNAME");
    struct hostent* host_info = gethostbyname(localhost);
    uint32_t ip_ = *reinterpret_cast<uint32_t*>(host_info->h_addr);
    uint16_t port_ = *reinterpret_cast<uint16_t*>(&serverport);
    m_id = net_id(ip_, port_);

    // Set net_id for message serial number stamping
    message::set_nid(m_id);

    // Name this postmaster's null synchronizer
    stringstream name;
    name << m_id << "null";
    null_name = name.str();
    nullsync = new sync_null_t(*this,null_name);

    // If this node is a client that is to connect to another postmaster...
    if(host)
    {
        int socket = connectbyportint(host, host_port);
        if(socket < 0)
        {
            stringstream err;
            err << "Host socket creation error: " << socket << "\n";
            throw err.str().c_str();
        }
        else
            rt_log << "Connected to host:port -> " << host << ":" << host_port << "\n";
        rtsocket* temp_sock = new rtsocket(socket);
        stringstream ss;

        ss << m_id;
        temp_sock->send(ss.str());
        ss.str("");

        // Request net_ids of all other postmasters in the system
        ss << RT_GET_NETIDS;
        temp_sock->send(ss.str());

        // Receive that information back
        temp_sock->receive(ss);

        hostent* address = gethostbyname(host);

        net_id tmp(*reinterpret_cast<uint32_t*>(address->h_addr_list[0]), host_port);

        postmasters[tmp] = temp_sock;

        // Create sockets for all the other postmasters
        if (ss.str() != "")
        {
            while((ss >> tmp))
            {
                socket = connectbyport_ip(tmp.ip(), tmp.port());
                if(socket < 0)
                {
                    stringstream err;
                    err << "Error creating socket to " << tmp.ip() << ":"
                        << tmp.port() << ": " << socket << endl;
                    throw err.str().c_str();
                }
                postmasters[tmp] = new rtsocket(socket);
            }
        }
    }
    // Set awaiting_postmasters so we know how many we have to get
    // population data from
    awaiting_postmasters = postmasters.size();

    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

    // Create the postmaster's polling (net_recv) thread
    int result = pthread_create(&recv_thread, &attr,
                                (void* (*)(void*)) start_polling_func,
                                (void*) this);
    if (result < 0)
        throw "Error creating net-receiving thread...\n";
}

/*
 * Destructor - deletes everything.
 * At this point, code has not been put through Valgrind or any other such
 * leak detector.  We don't see any noticable leaks during prolonged executions
 * but...
 */
postmaster::~postmaster()
{
    delete nullsync;
    delete gsync;

    close(server_socket);
    for(actormap::iterator it = actors.begin(); it != actors.end(); ++it)
        delete it->second;
    for(syncmap::iterator it = synchronizers.begin();
        it != synchronizers.end(); ++it)
    {
        delete it->second;
    }
    for(std::map<net_id, rtsocket*>::iterator it = postmasters.begin();
        it != postmasters.end(); ++it)
    {
        delete it->second;
    }
}

/*
 * Notify synchronizer s that a.m() ran and completed at t
 */
void postmaster::sm_notify_action(const syncref& s, const actorref& a,
                        const method& m, const rttime_t& t)
{
    // Ideally, this function would be refined to take a list of notifications
    // and dispatch them according to network layout (i.e. send one
    // consolidated notification per remote postmaster involved); this would
    // require the actors to form notification batches.

    std::map<syncref,actionlistener*>::iterator it = listeners.find(s);
    if (it == listeners.end())
    {
        rem_sync_map::iterator rem_it = rem_syncs.find(s);
        if (rem_it == rem_syncs.end())
        {
            rt_log << "sm_notify_action: did not find synchronizer local or remote\n";
        }
        else
        {
            std::stringstream ss;
            // message must go last in transmissions to ensure that params
            // get de-serialized properly
            ss << RT_NOTIF << " " << s << " " << a << " " << t << " " << m;
            postmasters[rem_it->second]->send(ss.str());
        }
    }
    else
        it->second->notify_action(a, m, t);
}

/*
 * Updates local method map to indicate that a.m() ran at t.  Note that
 * this is an area for future improvement.  Currently, the protocol is such
 * that every postmaster has a list of ALL methods in the system and there
 * is an excess of superfluous traffic updating lastrun and enabled/disabled
 * flags for methods.
 */
void postmaster::update_run(const actorref& a, const method& m,
                            const rttime_t& t)
{
    std::pair<actorref,method> p(a,m);
    methods.lock();
    methods[p].lastrun = t;
    methods.unlock();
}

/*
 * Update the lastrun of a local method and then send an RT_TIMEUPDATE to
 * all other postmasters so they may do the same.  See note above.
 */
void postmaster::update_local_run(const actorref& a, const method& m)
{
    rttime_t t = rt_gettime();
    update_run(a, m, t);

    for(std::map<net_id, rtsocket*>::iterator it = postmasters.begin();
        it != postmasters.end(); ++it)
    {
        std::stringstream ss;
        ss << RT_TIMEUPDATE << " " << a << " " << m << " " << t;
        it->second->send(ss.str());
    }
}

// void postmaster::sm_receive_msg(const syncref& s, const message& m)
// {
//     syncmap::iterator it = synchronizers.find(s);
//     if(it != synchronizers.end())
//         it->second->receive_msg(m);
//     else
//     {
//         rem_sync_map::iterator rem_it = rem_syncs.find(s);
//         if(rem_it == rem_syncs.end())
//             rt_log << "Cannot deliver message " << m.serial
//                    << " to " << s << ": " << s << "not found\n";
//         else
//         {
//             std::stringstream ss;
//             ss << RT_RECVMSG << " " << s << " " << m;
//             postmasters[rem_it->second]->send(ss.str());
//         }
//     }
// }

/*
 * This function is called by an actor which has previously received a message
 * registration to retrieve the message that it was notified about from the 
 * synchronizer.
 *
 * If the synchronizer is local to the actor then the message is simply
 * returned without delay.
 *
 * If the synchronizer is remote, then the actor will block and wait for a
 * signal of a condition variable before proceeding.
 */
message postmaster::sm_request_msg(const syncref& s, const actorref& a)
{
    syncmap::iterator it = synchronizers.find(s);
    if (it == synchronizers.end())
    {
        rem_sync_map::iterator rem_it = rem_syncs.find(s);
        if (rem_it == rem_syncs.end())
        {
            rt_log << "sm_request_msg: cannot find synchronizer local or remote\n";
            _exit(1);
            return null_msg;
        }
        else
        {
            using namespace boost;

            ret_param bp;
            get<1>(bp) = false;
            pthread_cond_init(&get<0>(bp), 0);

            // Insert blocked parameter
            recv_msgs.lock();
            recv_msgs[a] = bp;
            ret_param& mp = recv_msgs[a];
            recv_msgs.unlock();

            pthread_mutex_t mut = PTHREAD_MUTEX_INITIALIZER;

            std::stringstream ss;
            ss << RT_REQUEST << " " << s << " " << a;
            postmasters[rem_it->second]->send(ss.str());

            // Now wait until recipient signals completion
            pthread_mutex_lock(&mut);
            while(!get<1>(mp))
                pthread_cond_wait(&get<0>(mp), &mut);
            pthread_mutex_unlock(&mut);

            pthread_mutex_destroy(&mut);
            pthread_cond_destroy(&get<0>(mp));

            message msg(get<2>(mp));

            recv_msgs.lock();
            recv_msgs.erase(a);
            recv_msgs.unlock();

            return msg;
        }
    }
    else
    {
        return synchronizers[s]->request_msg(a);
    }
}

/*
 * Message to synchronizer to indicate that an actor has received a message
 * that it had requested.
 *
 * The message has not been acted upon at the time this call is made.  It is
 * simply part of the handshake to indicate that the message was received
 * properly.  Currently, functionality has been removed from this section. All
 * that the synchronizer does is to log the message.  We left it in place
 * as a hook for future actions.
 */
void postmaster::sm_ack_msg(const syncref& s, const message& m)
{
    syncmap::iterator it = synchronizers.find(s);
    if(it != synchronizers.end())
        it->second->ack_msg(m);
    else
    {
        rem_sync_map::iterator rem_it = rem_syncs.find(s);
        if(rem_it == rem_syncs.end())
            rt_log << "Cannot acknowledge message " << m.serial
                   << " to " << s << ": " << s << "not found\n";
        else
        {
            std::stringstream ss;
            ss << RT_ACKMSG << " " << s << " " << m;
            postmasters[rem_it->second]->send(ss.str());
        }
    }
}

// Quick function to make pairs as we use them frequently
template <class S, class T> std::pair<S, T> makepair(const S& s, const T& t)
{
    return std::pair<S, T>(s, t);
}

// Check to see if method a.m() is enabled
bool postmaster::am_enabled(const actorref& a, const method& m)
{
    bool b;
    methods.lock();
    method_map::iterator it = methods.find(makepair(a,m));
    if (it != methods.end())
        b = it -> second.enabled;
    else b = false;
    methods.unlock();
    return b;
}

/*
 * Set the enabled flag for a.m() to true.  If a is a local actor then let
 * all of the other postmasters know so they can update their maps.  If it is
 * a remote actor, then tell them to enable and they will tell everyone else.
 *
 * Bottom-line rule:  it's always the postmaster local to the actor that lets
 * all of the other postmasters know.
 */
void postmaster::am_enable(const actorref& a, const method& m)
{
    actormap::iterator it = actors.find(a);
    if(it != actors.end())
    {
        methods.lock();
        methods[makepair(a,m)].enabled = true;
        methods.unlock();
        std::stringstream ss;
        ss << RT_POST_ENABLE << " " << a << " " << m;
        for (std::map<net_id, rtsocket*>::iterator it = postmasters.begin();
             it != postmasters.end(); ++it)
        {
            it->second->send(ss.str());
        }
    }
    else // a is not local
    {
        rem_actor_map::iterator rem_it = rem_actors.find(a);
        if(rem_it == rem_actors.end())  // oh no, a is not remote either
            rt_log << "Cannot enable" << m << ": Actor "
                   << a << " not found\n";
        else
        {
            std::stringstream ss;
            ss << RT_ENABLE << " " << a << " " << m;
            postmasters[rem_it->second]->send(ss.str());
        }
    }
}

/*
 * Function called by the KILLNODE return value of the exception handler
 * contained within synchronizers.  This function sends a communication to
 * all other postmasters to inform them that this node is going down, and
 * then throws an exception.
 *
 * Currently, it is the node where the synchronizer that caught the exception
 * resides that is terminated.
 *
 * Future plan:  It makes more sense that the node to terminate should be
 * the one that contained the destination actor from the message that caused
 * the exception.
 */
void postmaster::terminate_node()
{
    for (std::map<net_id, rtsocket*>::iterator it = postmasters.begin();
         it != postmasters.end(); ++it)
    {
        std::stringstream ss;
        ss << RT_TERMINATE;
        it->second->send(ss.str());
    }
    rt_log << "Exception caused termination of this node...goodbye" << std::endl;
    throw "Terminated by exception";
}

/*
 * Function called by the TERMINATE return value of the exception handler
 * contained within synchronizers.  This function sends a communication to
 * all other postmasters to inform them that they should all terminate then
 * it throws an exception to terminate itself.
 */
void postmaster::terminate_system()
{
    for (std::map<net_id, rtsocket*>::iterator it = postmasters.begin();
         it != postmasters.end(); ++it)
    {
        std::stringstream ss;
        ss << RT_KILLALL;
        it->second->send(ss.str());
    }
    rt_log << "Exception caused terminate system...goodbye" << std::endl;
    throw "Terminated by exception";
}

/*
 * Set the enabled flag for a.m() to false.  If a is a local actor then let
 * all of the other postmasters know so they can update their maps.  If it is
 * a remote actor, then tell the remote actor's postmaster to disable and
 * tell everyone else.
 *
 * Bottom-line rule:  it's always the postmaster local to the actor that lets
 * all of the other postmasters know.
 */
void postmaster::am_disable(const actorref& a, const method& m)
{
    actormap::iterator it = actors.find(a);
    if(it != actors.end())
    {
        methods.lock();
        methods[makepair(a,m)].enabled = false;
        methods.unlock();
        std::stringstream ss;
        ss << RT_POST_DISABLE << " " << a << " " << m;
        for (std::map<net_id, rtsocket*>::iterator it = postmasters.begin();
             it != postmasters.end(); ++it)
        {
            it->second->send(ss.str());
        }
    }
    else
    {
        rem_actor_map::iterator rem_it = rem_actors.find(a);
        if(rem_it == rem_actors.end())
            rt_log << "Cannot disable" << m << ": Actor "
                   << a << " not found\n";
        else
        {
            std::stringstream ss;
            ss << RT_DISABLE << " " << a << " " << m;
            postmasters[rem_it->second]->send(ss.str());
        }
    }
}

/*
 * Return the lastrun time of a method.  This function can be called locally
 * by all postmasters as they all contain complete method maps with this
 * information.
 *
 * It is obvious that updating the method maps generates a great deal of
 * extra communication, and this is a planned fix in the future.
 */
rttime_t postmaster::am_lastrun(const actorref& a, const method& m)
{
    methods.lock();
    rttime_t lr = methods[std::pair<actorref,method>(a,m)].lastrun;
    methods.unlock();
    return lr;
}

/*
 * This function allows an actor to register its methods with the postmaster
 * during its instantiation.  (Method list is prepared during code generation
 * in the codevisitor.)
 */
void postmaster::reg_methods(const actorref& a, const std::list<method>& ml)
{
    for (std::list<method>::const_iterator it = ml.begin(); it != ml.end(); ++it)
    {
        std::pair<actorref,method> p(a,*it);
        methods.lock();
        methods[p] = method_info();
        methods.unlock();
    }
}

/*
 * Function to register a message with an actor - part of the message handshake
 * specified in docs/message_handshake.
 *
 * Note: this function does not send a message to an actor, it merely lets the
 *       actor know that a synchronizer has a message for it
 *
 * If the funtion finds the destination actor locally, it just calls its
 * register_msg().
 *
 * Else, if it cannot find the destination actor remotely, it caches the
 * registration in its local msg_reg_cache (locked list).
 *
 * Else, if it does find the dest actor (remote) then it sends the RT_REGMSG to
 * it.
 */
void postmaster::am_register_msg(const actorref& a, syncref s, rttime_t priority)
{
    actormap::iterator it = actors.find(a);
    if(it != actors.end())
        it->second->register_msg(s, priority);
    else
    {
        rem_actor_map::iterator rem_it = rem_actors.find(a);
        if(rem_it == rem_actors.end())
        {
            msg_reg_cache.lock();
            msg_reg_cache.push_back(boost::make_tuple(a, s, priority));
            msg_reg_cache.unlock();
            rt_log << "Cannot register message: Actor " << a
                   << " not found (cached)\n";
        }
        else
        {
            std::stringstream ss;
            ss << RT_REGMSG << " " << a << " " << s << " " << priority;
            postmasters[rem_it->second]->send(ss.str());
        }
    }
}

/*
 * Method for listeners to be registered with actors (so they know who to
 * notify when they complete tasks).
 *
 * If the actor does not exist at the time of the call, then the actor/sync
 * pair is put into the listener_cache structure which will be checked whenever
 * new actors enter the system.
 */
void postmaster::am_register_listener(const actorref& a, const syncref& s)
{
    actormap::iterator it = actors.find(a);
    if(it != actors.end())
       it->second->register_listener(s);
    else
    {
        rem_actor_map::iterator rem_it = rem_actors.find(a);
        if(rem_it == rem_actors.end())
        {
            listener_cache.lock();
            listener_cache.push_back(reg_pair(a, s));
            listener_cache.unlock();
            rt_log << "Cannot register listener: Actor "
                   << a << " not found (cached)\n";
        }
        else
        {
            std::stringstream ss;
            ss << RT_REG_LISTENER << " " << a << " " << s;
            postmasters[rem_it->second]->send(ss.str());
        }
    }
}

/*
 * Function for an actor to register with its postmaster
 */ 
void postmaster::reg(const actorref& r, actor* a)
{
    rt_log << "Registering " << r << "\n";
    actors[r] = a;
}

/*
 * Function for a synchronizer to register with its postmaster
 */ 
void postmaster::reg(const syncref& r, synchronizer* s)
{
    rt_log << "Registering " << r << "\n";
    listeners[r] = synchronizers[r] = s;
}

/*
 * Function for the local global synchronizer to register with the postmaster
 */ 
void postmaster::reg(const syncref& r, globalsync* s)
{
    rt_log << "Registering " << r << "\n";
    listeners[r] = gsync = s;
}

/*
 * Function to register a synchronizer and its synchronized pair with the
 * postmaster
 */ 
void postmaster::reg(const syncref& s, const actorref& a1, const actorref& a2)
{
    rt_log << "Registering " << s << " as synchronizing " << a1 << " and "
           << a2 << "\n";
    // NB: If s synchronizes a1->a2, it also synchronizes a2->a1
    typedef std::pair<actorref, actorref> apair;
    apair p1(a1, a2), p2(a2, a1);
    sync_rel[p1] = s;
    sync_rel[p2] = s;
}

/*
 * Function to send a message.  It first checks to see if the source and 
 * destination actor have a synchronizer registered for them.  If not, it sends
 * the message to the local null synchronizer as an unsynchronized message.
 *
 * If it found a registered synchronzier, it then checks to see if that sync
 * is local; if so, sends message to it.
 *
 * If not local, it checks its list of remote synchronizers (and it *will* find
 * one), it sends to the appropriate remote one
 */
void postmaster::sendmsg(const message& m)
{
    std::pair<actorref, actorref> p(m.sender, m.receiver);
    sync_rel_map::iterator it = sync_rel.find(p);

    if(it == sync_rel.end())
    {
        std::stringstream ss;
        ss << m_id << "null";
        rt_log << "Passing message " << m.serial << " to null synchronizer\n";
        // Feed it to the null synchronizer
        synchronizers[ss.str()]->receive_msg(m);
    }
    else if(synchronizers.find(it->second) != synchronizers.end())
    {
        rt_log << "Passing to synchronizer" << "\n";
        synchronizers[it->second]->receive_msg(m);
    }
    else if(rem_syncs.find(it->second) != rem_syncs.end())
    {
        std::stringstream ss;
        ss << RT_MESSAGE << " " << m;
        postmasters[rem_syncs[it->second]]->send(ss.str());
    }
}

/*
 * Function to send a synchronous message.  A synchronous message in the
 * context of rtsynchronizers implies that the originating actor blocks
 * until the destination actor signals that it has completed execution of
 * the function specified in the message.
 
 */
void postmaster::sendsyncmsg(const message& m)
{
    using namespace boost;
    block_param bp;
    get<1>(bp) = false;
    pthread_cond_init(&get<0>(bp), 0);
    pthread_mutex_t mut = PTHREAD_MUTEX_INITIALIZER;

    // Insert blocked parameter
    return_map.lock();
    return_map[m.serial] = bp;
    block_param& mp = return_map[m.serial];
    return_map.unlock();

    sendmsg(m);

    // Now wait until recipient signals completion
    pthread_mutex_lock(&mut);
    while(!get<1>(mp))
        pthread_cond_wait(&get<0>(mp), &mut);
    pthread_mutex_unlock(&mut);

    pthread_mutex_destroy(&mut);
    pthread_cond_destroy(&get<0>(mp));

    return_map.lock();
    return_map.erase(m.serial);
    return_map.unlock();
}

/*
 * Allows destination actor of a synchronous message to respond back to source
 * actor (by way of condition variable signalling) to tell it that it has
 * completed execution of the message, thereby allowing the blocked source actor
 * to continue execution.
 */
void postmaster::recvsyncmsg(const message& m)
{
    using namespace boost;

    if(actors.find(m.sender) == actors.end()) // sender is remote
    {
        std::stringstream ss;
        ss << RT_SYNCNOTE << " " << m;
        rem_actor_map::iterator it = rem_actors.find(m.sender);
        if(it == rem_actors.end())
        {
            rt_log << "ERROR: Cannot relay synchronous message response; "
                   << "original sender not found\n";
        }
        else
        {
            postmasters[it->second]->send(ss.str());
        }
    }
    else
    {
        return_map.lock();
        block_param& bp = return_map[m.serial];
        return_map.unlock();

        // Once a synchronous call has completeted, signal the waiting thread
        get<1>(bp) = true;
        pthread_cond_signal(&get<0>(bp));
    }
}

// Run method which actually starts execution my sending a distinct message.
void postmaster::run(const actorref& a, const method& m, bool connect)
{
    using namespace std;

#ifdef USE_SCHED
    sched_begin();
#endif

    for(list<reg_pair>::iterator it = reg_requests.begin();
        it != reg_requests.end(); ++it)
    {
        actors[it->first]->register_listener(it->second);
    }

    if(connect)
    {
        // Get data from the other postmasters
        for(map<net_id, rtsocket*>::iterator it = postmasters.begin();
            it != postmasters.end(); ++it)
        {
            std::stringstream ss;
            ss << m_id;
            it->second->send(ss.str());
            ss.str("");
            ss << RT_POPULATE;
            it->second->send(ss.str());
            send_population_data(it->first);
        }
    }

    cout << "Running program" << endl;
    message msg("PostMaster", "run", a, m, params());
    stringstream syncname;
    syncname << m_id << "null";
    msg.sync = syncname.str();

    pollfd p = { 0, POLLERR, POLLERR };

    // wait until we've heard from all other known postmasters before beginning
    while (awaiting_postmasters != 0)
        poll(&p, 1, 250);
        
    // begin
    actors[a]->execute(msg);

    // Never return
    while(true)
        poll(&p, 1, -1);
}

// run method that does not begin execution by itself.
void postmaster::run(bool connect)
{
    using namespace std;

#ifdef USE_SCHED
    sched_begin();
#endif

    for(list<reg_pair>::iterator it = reg_requests.begin();
        it != reg_requests.end(); ++it)
    {
        actors[it->first]->register_listener(it->second);
    }

    rt_log << "Waiting for other nodes before execution." << endl;

    if(connect)
    {
        // Get data from the other postmasters
        for(map<net_id, rtsocket*>::iterator it = postmasters.begin();
            it != postmasters.end(); ++it)
        {
            std::stringstream ss;
            ss << m_id;
            it->second->send(ss.str());
            ss.str("");
            ss << RT_POPULATE;
            it->second->send(ss.str());
            send_population_data(it->first);
        }
    }

    // Never return
    pollfd p = { 0, POLLERR, POLLERR };
    while(true)
        poll(&p, 1, -1);
}

/*
 * This function receives population data from another postmaster and parses
 * it into its constituent parts, inserting them into the appropriate data
 * structures.
 *
 * See the send_population_data() function comments to see what a population
 * packet consists of.  Each data section is separated by $.
 *
 * The final two sections of this function check through the two caches that
 * the postmaster maintains to deal with entities which have been referenced
 * but do not exist yet.
 *
 * The first of these is the listener cache and the second is the message
 * registration cache (which holds notifications of messages destined for 
 * actors which did not exist at the time of sending and have not thrown
 * timing exceptions)
 */
void postmaster::populate(std::stringstream& ss, net_id n)
{
    std::string s;
    while ((ss >> s) && (s != "$"))
        rem_actors[s] = n;

    while ((ss >> s) && (s != "$"))
        rem_syncs[s] = n;

    while ((ss >> s) && (s != "$"))
    {
        std::pair<actorref, actorref> ap;
        syncref sr;
        ap.first = s;
        ss >> ap.second >> sr;
        sync_rel[ap] = sr;
    }
    while ((ss >> s) && (s != "$"))
    {
        std::pair<actorref, method> ap;
        bool b;
        ap.first = s;
        ss >> ap.second >> b;
        methods.lock();
        methods[ap].enabled = b;
        methods.unlock();
    }
    // cache of listeners
    listener_cache.lock();
    for(std::list<reg_pair>::iterator it = listener_cache.begin();
        it != listener_cache.end(); )
    {
        if(rem_actors.find(it->first) != rem_actors.end())
        {
            listener_cache.unlock();
            am_register_listener(it->first, it->second);
            listener_cache.lock();
            it = listener_cache.erase(it);
        }
        else
            ++it;
    }
    listener_cache.unlock();

    using namespace boost;

    msg_reg_cache.lock();
    for(std::list<tuple<actorref, syncref, rttime_t> >::iterator it
          = msg_reg_cache.begin(); it != msg_reg_cache.end(); )
    {
        if(get<2>(*it) < rt_gettime())
            it = msg_reg_cache.erase(it);
        else if(rem_actors.find(get<0>(*it)) != rem_actors.end())
        {
            tuple<actorref, syncref, rttime_t> tup = *it;
            it = msg_reg_cache.erase(it);
            msg_reg_cache.unlock();
            am_register_msg(get<0>(tup), get<1>(tup), get<2>(tup));
            msg_reg_cache.lock();
        }
        else
            ++it;
    }
    msg_reg_cache.unlock();
    awaiting_postmasters--;
}

bool operator<(const postmaster::rem_actor_map::iterator& i1,
               const postmaster::rem_actor_map::iterator& i2)
{
    return i1->second < i2->second;
}

/*
 * This function is utilized when one postmaster is about to terminate.
 * It sends a communication to every other postmaster and this is the code
 * that they execute.  The purpose:  to remove all traces of the terminating
 * postmaster from the data structures of the postmasters that still exist.
 *
 * It obtains locks on all of the critical data structures at the beginning
 * of the function to ensure that no one else can do anything until this
 * operation is complete.
 */
void postmaster::terminate(const net_id& nid)
{
    std::set<rem_actor_map::iterator> actors;

    postmasters.lock();
    rem_actors.lock();
    rem_syncs.lock();
    methods.lock();
    recv_msgs.lock();

    // Remove the methods that are local to the terminating postmaster
    for (method_map::iterator it = methods.begin(); it != methods.end();
         ++it)
    {
        const actorref& a = it->first.first;
        rem_actor_map::iterator ra_it = rem_actors.find(a);
        if (ra_it != rem_actors.end() && ra_it->second == nid)
        {
            actors.insert(ra_it);
            methods.erase(it);
        }
    }

    // Remove the actors
    for (std::set<rem_actor_map::iterator>::iterator it = actors.begin();
         it != actors.end(); ++it)
    {
        recv_msg_map::iterator rm_it = recv_msgs.find((*it)->first);
        if (rm_it != recv_msgs.end())
            recv_msgs.erase(rm_it);
        rem_actors.erase(*it);
    }

    // Remove the synchronizers
    for (rem_sync_map::iterator it = rem_syncs.begin(); it != rem_syncs.end();
         ++it)
    {
        if (it->second == nid)
            rem_syncs.erase(it);
    }

    // Remove the reference to the postmaster that is disappearing
    for (std::map<net_id, rtsocket*>::iterator it = postmasters.begin();
         it != postmasters.end(); ++it)
    {
        if (it->first == nid)
            postmasters.erase(it);
    }

    recv_msgs.unlock();
    methods.unlock();
    rem_syncs.unlock();
    rem_actors.unlock();
    postmasters.unlock();
}

/*
 * This function sends the calling postmasters population data to another
 * postmaster (pointed to by n).
 *
 * Population data consists of:
 *   all local actors
 *   all local synchronizers
 *   all local coordination relationships (actor,actor,sync)
 *   all methods in the method map (local and remote)
 *
 * This communication only occurs when a new postmaster joins a system
 * already in existence.  Once the new postmaster learns who the other
 * postmasters are, it sends a request for this population data to each
 * of them.
 */
void postmaster::send_population_data(net_id n)
{
    std::stringstream ss;
    ss << RT_POPULATION << " ";
    // actors
    for (actormap::iterator it = actors.begin(); it != actors.end();
         ++it)
    {
        ss << it->first << " ";
    }
    ss << "$ ";
    // synchronizers
    for (syncmap::iterator it = synchronizers.begin();
         it != synchronizers.end(); ++it)
    {
        ss << it->first << " ";
    }
    ss << "$ ";
    // coordination relationships
    for(sync_rel_map::iterator it = sync_rel.begin(); it != sync_rel.end();
        ++it)
    {
        ss << it->first.first << " " << it->first.second << " "
           << it->second << " ";
    }
    ss << "$ ";
    // methods
    methods.lock();
    for(method_map::iterator it = methods.begin(); it != methods.end(); ++it)
    {
        ss << it->first.first << " " << it->first.second << " "
           << it->second.enabled << " ";
    }
    methods.unlock();
    ss << "$ ";

    rtsocket* r = postmasters[n];
    r->send(ss.str());
}

/*
 * This function is the master loop run by the postmasters polling thread
 *
 * Handles all net reception traffic to this postmaster.  Monitors server socket
 * and each of the sockets formed between the other postmasters.
 */
void postmaster::net_recv()
{
    using namespace std;
    vector<net_id> nids;
    for(;;)
    {
        postmasters.lock();
        int pfdcount = postmasters.size()+1;
        pollfd* pfds = new pollfd[pfdcount];
        nids.clear();
        int count = 0;
        
        // populate pollfd structure array for this run
        for(map<net_id, rtsocket*>::iterator it = postmasters.begin();
            it != postmasters.end(); ++it, ++count)
        {
            pfds[count].fd = it->second->getsd();
            pfds[count].events = POLLIN;
            nids.push_back(it->first);
        }
        postmasters.unlock();
        pfds[count].fd = server_socket;
        pfds[count].events = POLLIN;

        // wait until we see activity on a socket
        poll(pfds, pfdcount, -1);

        count = 0;
        postmasters.lock();
        
        // determine which socket we saw activity on and call parse_data
        for(map<net_id, rtsocket*>::iterator it = postmasters.begin();
            it != postmasters.end(); ++it, ++count)
        {

            // Do we have incoming data on this socket?
            if((pfds[count].revents == POLLIN or pfds[count].revents == POLLPRI)
               and postmasters.find(nids[count]) != postmasters.end())
            {
                parse_data(it->first, *(it->second));
            }
        }
        postmasters.unlock();

        // check to see if there was an event on the server socket
        if(pfds[count].revents == POLLIN or pfds[count].revents == POLLPRI)
        {
            sockaddr_in sa;
            socklen_t sl;
            int new_sock = accept(server_socket, (sockaddr*)&sa, &sl);
            rtsocket* temp = new rtsocket(new_sock);
            stringstream ss;
            temp->receive(ss);
            net_id nid;
            ss >> nid;

            postmasters.lock();
            postmasters[nid] = temp;
            postmasters.unlock();
        }

        delete[] pfds;
    }
}

/*
 * This function is called by net_recv() when it has identified that there
 * is a communication inbound on a socket.  The parameter nid is the net_id
 * of the postmaster the communication is being received from and sock is the
 * rtsocket we use for communication.
 *
 * The first action is to receive that data packet and put it into a
 * stringstream.  The first value in the packet will be an int which represents
 * the command that this message conveys.
 *
 * The case essentially acts as a jump table to then process the information
 * received in the rest of the packet.
 *
 * For a description of the message_type values, see the comm_protocol
 * document in the docs directory.
 *
 * One key weakness in this design is that we keep a lock on the postmasters
 * map from the net_recv function all the way through parse data and any calls
 * we make from here.  It is likely that there would be a performance
 * improvement if this is remedied.
 */
void postmaster::parse_data(const net_id& nid, rtsocket& sock)
{
    std::stringstream ss;
    sock.receive(ss);
    int type;
    ss >> type;

    syncref s;
    actorref a;
    switch(static_cast<message_type>(type))
    {
        case RT_MESSAGE:
        {
            message m = null_msg;
            ss >> m;
            sendmsg(m);
        } break;
        case RT_NOTIF:
        {
            method m;
            rttime_t t;
            ss >> s >> a >> t >> m;
            sm_notify_action(s, a, m, t);
        } break;
        case RT_REQUEST:
        {
            ss >> s >> a;
            message m = sm_request_msg(s, a);
            std::stringstream buf;
            buf << RT_REQRESPONSE << " " << m;
            sock.send(buf.str());
        } break;
        case RT_REQRESPONSE:
        {
            using namespace boost;
            message m = null_msg;
            ss >> m;
            recv_msgs.lock();
            ret_param& ret = recv_msgs[m.receiver];
            pthread_cond_t& cond = get<0>(ret);
            recv_msgs[m.receiver] = boost::make_tuple(cond, true, m);
            recv_msgs.unlock();
            pthread_cond_signal(&cond);
        } break;
        case RT_REGMSG:
        {
            rttime_t t;
            ss >> a >> s >> t;
            am_register_msg(a, s, t);
        } break;
        case RT_ENABLE:
        {
            method m;
            ss >> a >> m;
            am_enable(a, m);
        } break;
        case RT_DISABLE:
        {
            method m;
            ss >> a >> m;
            am_disable(a, m);
        } break;
        case RT_REG_LISTENER:
        {
            ss >> a >> s;
            am_register_listener(a, s);
        } break;
/*        case RT_RECVMSG:
        {
            message m = null_msg;
            ss >> s >> m;
            sm_receive_msg(s, m);
        } break;*/
        case RT_ACKMSG:
        {
            message m = null_msg;
            ss >> s >> m;
            sm_ack_msg(s, m);
        } break;
        case RT_POST_ENABLE:
        {
            method m;
            ss >> a >> m;
            methods.lock();
            methods[makepair(a,m)].enabled = true;
            methods.unlock();
        } break;
        case RT_POST_DISABLE:
        {
            method m;
            ss >> a >> m;
            methods.lock();
            methods[makepair(a,m)].enabled = false;
            methods.unlock();
        } break;
        case RT_GET_NETIDS:
        {
            std::stringstream buf;
            // We already have a mutex lock on postmasters
            for(std::map<net_id, rtsocket*>::iterator it = postmasters.begin();
                it != postmasters.end(); ++it)
            {
                if (it->first != nid)
                    buf << it->first << " ";
            }
            sock.send(buf.str());
        } break;
        case RT_POPULATE:
        {
            send_population_data(nid);
        } break;
        case RT_POPULATION:
        {
            populate(ss,nid);
        } break;
        case RT_TERMINATE:
        {
            terminate(nid);
        } break;
        case RT_KILLALL:
        {
            rt_log << "Exception caused terminate system...goodbye"
                   << std::endl;
            throw "Terminated by exception";
        } break;
        case RT_TIMEUPDATE:
        {
            method m;
            rttime_t t;
            ss >> a >> m >> t;
            update_run(a, m, t);
        } break;
        case RT_SYNCNOTE:
        {
            message msg;
            ss >> msg;
            recvsyncmsg(msg);
        } break;
        default:
            break;
    }
}
