/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef POSTMASTER_H
#define POSTMASTER_H

#include <utility>
#include <pthread.h>
#include <iostream>

#include <boost/tuple/tuple.hpp>

#include "messages.h"
#include "net_id.h"
#include "mutex.h"
#include "timing.h"
#include "reftypes.h"

#include "lock_map.h"
#include "lock_list.h"

class actor;
class synchronizer;
class globalsync;
class rtsocket;
class actionlistener;


// Class for the node-wide PostMaster object that relays all messages
// between actors and synchronizers.

/*
 * The postmaster of a node serves both as a message relay unit and as
 * communication point between other postmasters (nodes) within a distributed
 * system.
 *
 * Most functions which act upon actors or synchronizers contain a large
 * IF structure which determines if the destination actor or synchronizer
 * is local or remote.  If it's local, it makes the call there and then.  If
 * it's remote, it sends the data to the destination postmaster which will then
 * run the same function and (of course) determine that the destination is
 * local.
 */
class postmaster
{
    // thread which will be used to receive data from other postmasters
    pthread_t recv_thread;

    struct method_info;

    // Fixed constants defeat the purpose of enums but ease debugging
    // Used for command signature for communication protocol
    enum message_type
    {
        RT_MESSAGE      = 1,
        RT_NOTIF        = 2,
        RT_REQUEST      = 3,
        RT_REQRESPONSE  = 4,
        RT_REGMSG       = 5,
        RT_ENABLE       = 6,
        RT_DISABLE      = 7,
        RT_REG_LISTENER = 8,
        //RT_RECVMSG      = 9,
        RT_ACKMSG       = 9,
        RT_POST_ENABLE  = 10,
        RT_POST_DISABLE = 11,
        RT_GET_NETIDS   = 12,
        RT_POPULATE     = 13,
        RT_POPULATION   = 14,
        RT_TERMINATE    = 15,
        RT_KILLALL      = 16,
        RT_TIMEUPDATE   = 17,
        RT_SYNCNOTE     = 18
    };

    // net_id of this postmaster
    net_id m_id;

    // name of the null synchronizer on this node
    std::string null_name;

    // server socket for this postmaster
    int server_socket;

    // How many postmasters are we waiting to hear from
    int awaiting_postmasters;

    // Track all other postmasters in the system
    lock_map<net_id, rtsocket*> postmasters;

    // Track actors
    typedef lock_map<actorref, actor*> actormap;
    actormap actors;

    // Track what synchronizers listen to what
    typedef lock_map<syncref, actionlistener*> listenermap;
    listenermap listeners;

    // Track synchronizers
    typedef lock_map<syncref, synchronizer*> syncmap;
    syncmap synchronizers;

    // Track what synchronizers guard what actor<->actor relation
    typedef lock_map<std::pair<actorref, actorref>, syncref> sync_rel_map;
    sync_rel_map sync_rel;

    // Global synchronizer
    globalsync* gsync;
    // Null synchronizer
    synchronizer* nullsync;

    typedef std::pair<actorref, syncref> reg_pair;
    lock_list<reg_pair> reg_requests;
    // Store references to listeners not yet in the system
    lock_list<reg_pair> listener_cache;
    // Cache message registrations for actors not yet in the system
    lock_list<boost::tuple<actorref, syncref, rttime_t> > msg_reg_cache;

    // Blocked parameters: Used for synchronous message passing
    typedef boost::tuple<pthread_cond_t, bool> block_param;
    // Blocked parameters by message serial number
    lock_map<message::serial_no, block_param> return_map;

    typedef boost::tuple<pthread_cond_t, bool, message> ret_param;
    //typedef std::map<actorref,std::list<ret_param> > recv_msg_map;
    typedef lock_map<actorref, ret_param> recv_msg_map;
    recv_msg_map recv_msgs;

    // Map to hold data about remote actors
    typedef lock_map<actorref,net_id> rem_actor_map;
    rem_actor_map rem_actors;
    friend bool operator<(const rem_actor_map::iterator& i1,
                          const rem_actor_map::iterator& i2);

    // Map to hold data about remote synchronizers
    typedef lock_map<syncref, net_id> rem_sync_map;
    rem_sync_map rem_syncs;

    // Map that holds information about all methods in system (local and
    // remote)
    typedef lock_map<std::pair<actorref, method>, method_info> method_map;
    method_map methods;

    // Function which runs as a thread to receive data from other postmasters
    void net_recv();

    // Handle population data from another postmaster and store info
    // into our structures
    void populate(std::stringstream& ss, net_id n);

    // Send data about this postmaster to others
    void send_population_data(net_id n);

    // Parse communication packet sent by someone else to us.
    void parse_data(const net_id& nid, rtsocket& sock);

    // Purge all data pertaining to another node that has terminated
    void terminate(const net_id& nid);

    // Non member function to start postmaster polling thread
    friend void* start_polling_func(postmaster* p);

    // Update the last run time of a method.
    void update_run(const actorref& a, const method& m,
                    const rttime_t& t);
public:
    postmaster(const char* host, unsigned short host_port,
               unsigned short serverport, int backlog=10);
    ~postmaster();

    // Kill current node and let all other postmasters know
    void terminate_node();
    // Kill all nodes in system
    void terminate_system();

    // Register entities with the postmaster
    void reg(const actorref& r, actor* a);
    void reg(const syncref& s, synchronizer* s);
    void reg(const syncref& s, globalsync* s);
    // Register a synchronizer as listening to the relation a1<->a2
    void reg(const syncref& s, const actorref& a1, const actorref& a2);

    // An actor broadcasts execution completion
    void broadcast_action(const actorref& a, const method& m);
    // Method for actors to send messages
    void sendmsg(const message& m);
    // Method for actors to send synchronous messages
    void sendsyncmsg(const message& m);
    // Method for actors to receive synchronous messages
    void recvsyncmsg(const message& m);

    // This function starts the system; it sends message m to actor a
    void run(const actorref& a, const method& m, bool connect);
    // Start a passive system: Since no initial message is passed, the
    // system must be activated by a message from outside
    void run(bool connect);

    // Actor notifies synchronizer of an event (i.e. method a.m() called)
    void sm_notify_action(const syncref& s, const actorref& a,
                          const method& m, const rttime_t& t);

/*    // Synchronizer receives message from actor
    void sm_receive_msg(const syncref& s, const message& m);*/

    // Actor requests the most urgent message pending at synchronizer
    message sm_request_msg(const syncref& s, const actorref& a);
    // Actor acknowledges the reception of a message from synchronizer
    void sm_ack_msg(const syncref& s, const message& m);

    // Check to see if a.m() is currently allowed to execute
    bool am_enabled(const actorref& a, const method& m);

    // Function which enables a method and notifies others
    void am_enable(const actorref& a, const method& m);
    // Function which disables a method and notifies others
    void am_disable(const actorref& a, const method& m);

    // Find out when a.m() was last run
    rttime_t am_lastrun(const actorref& a, const method& m);
    // Synchronizer s notifies actor a that it has a message waiting
    void am_register_msg(const actorref& a, syncref s, rttime_t priority);
    // Synchronizer s registers itself as a listener to a
    void am_register_listener(const actorref& a, const syncref& s);
    // Actor updates the local "last run" timestamp
    void update_local_run(const actorref& a, const method& m);
    // Actor notifies postmaster of what methods it contains
    void reg_methods(const actorref& a, const std::list<method>& ml);

    // Returns the net_id of the postmaster
    net_id get_nid() { return m_id; }

};

// Keeps information that we want to know about all methods
struct postmaster::method_info
{
    bool enabled;
    rttime_t lastrun;

    method_info() : enabled(true), lastrun(rt_startup) { }
};


#endif // #define POSTMASTER_H
