/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef RTSOCKET_H
#define RTSOCKET_H

#include "mutex.h"
#include <sys/socket.h>
#include <sys/types.h>
#include <sstream>

/*
 * Class representing a socket and providing wrappers for the system
 * send and recv functions.  Mutex protected to ensure communication
 * attempts are never split.
 */
class rtsocket : public mutex
{
    int m_sd;   // socket descriptor

    rtsocket(const rtsocket&);
    const rtsocket& operator=(const rtsocket&);
public:
    explicit rtsocket(int sd) : m_sd(sd) { }

    ~rtsocket() {
        close(m_sd);
    }

    // Function to send data through socket.  Appends size of data portion to
    // head of data to conform with protocol developed for this program.
    // Data to send is passed-in through s
    void send(const std::string& s) {
        std::stringstream ss;
        ss << s.size()+1 << " " << s;
        lock();
        ::send(m_sd, ss.str().c_str(), ss.str().size()+1, 0);
        unlock();
    }

    // Function to receive data from the socket.  Reads in size of expected
    // packet and then continues to read in that much information.
    // Stores data in the stringstream passed in.
    void receive(std::stringstream& buf) {
        using namespace std;
        stringstream ss;
        buf.str("");

        // Read in the size
        for(;;)
        {
            char c;
            recv(m_sd, &c, 1, 0);
            if(c == ' ')
                break;
            ss << c;
        }
        int msglen;
        ss >> msglen;

        if(msglen == 1)
        {
            char c;
            recv(m_sd, &c, 1, 0);
            return;
        }
        else if(msglen < 1)
            return;

        char* buffer = new char[msglen];
        char* bufp = buffer;
        // Not using C-string -- we don't -want- the null byte
        int msg_rem = msglen-1;
        while(msg_rem)
        {
            int n = recv(m_sd, bufp, msg_rem, 0);
            msg_rem -= n;
            buf.write(bufp, n);
            bufp += n;
        }
        recv(m_sd, bufp++, 1, 0);
    }

    // Return the socket descriptor
    int getsd() { return m_sd; }
};

#endif // #define RTSOCKET_H
