/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <deque>
#include <limits>
#include <pthread.h>
#include <semaphore.h>
#include <signal.h>
#include <sys/types.h>
#include <unistd.h>
#include <boost/cstdint.hpp>
#include "schedulable.h"
#include "timing.h"
#include "logger.h"

// map keyed by pthread_t to look up rt_sched_sync* which because synchronizers
// inherit from them, contain the condition variables that the sync will lock
// on in the signal handler
std::map<pthread_t, rt_sched_sync*> cond_table_sync;
// same as above but for actors
std::map<pthread_t, rt_sched_actor*> cond_table_actor;

/* Signal handler to catch signals sent to actor schedulable entities.
 * Pre: the actor thread must be running before being sent a signal.  If
 *      this is not the case, we will call this function from within this
 *      function, which may have unpredictable results.
 * The thread will sit here waiting on the pthread_cond_signal to leave
 * and continue execution
 * Post: the actor thread will continue running at the location in its
 *       code where it left to handle the signal.
 */
void rt_sighandler_actor(int sig)
{
    static pthread_mutex_t cond_mutex = PTHREAD_MUTEX_INITIALIZER;
    pthread_mutex_lock(&cond_mutex);
    rt_sched_actor* self = cond_table_actor[pthread_self()];
    while(!self->rt_sched_bool)
        pthread_cond_wait(&self->rt_sched_token, &cond_mutex);
    pthread_mutex_unlock(&cond_mutex);
}

/* Signal handler to catch signals sent to synchronizer schedulable entities.
 * Pre: the sync thread must be running before being sent a signal.  If
 *      this is not the case, we will call this function from within this
 *      function, which may have unpredictable results.
 * The thread will sit here waiting on the pthread_cond_signal to leave
 * and continue execution
 * Post: the sync thread will continue running at the location in its
 *       code where it left to handle the signal.
 */
void rt_sighandler_sync(int sig)
{
    static pthread_mutex_t cond_mutex = PTHREAD_MUTEX_INITIALIZER;
    pthread_mutex_lock(&cond_mutex);
    rt_sched_sync* self = cond_table_sync[pthread_self()];
    while(!self->rt_sched_bool)
        pthread_cond_wait(&self->rt_sched_token, &cond_mutex);
    pthread_mutex_unlock(&cond_mutex);
}

// used to store actor threads in the deque of deques
typedef std::pair<const pthread_t*, rt_sched_actor*> sched_entity_actor;
// used to store sync threads in the vector of synchronizers
typedef std::pair<const pthread_t*, rt_sched_sync*> sched_entity_sync;

// Source: http://www.gamedev.net/community/forums/topic.asp?topic_id=255581
// Fast base-2 logarithm approximation. Note: Does -not- work for log2(0)!
inline uint32_t log2(uint32_t n)
{
    *((float*)&n) = (float)n;
    return ((n & (~((1<<23) - 1))) >> 23) - 127;
}

/* The function that represents the scheduler.
 * General Operation:
 *   while(1)
 *     calculate new actor priority levels
 *     stop synchronizers that have been running
 *     start actors of highest priority level
 *       wait actor_quantum * number of actors started
 *     stop actors that were started
 *     start all synchronizers
 *       wait sync_quantum * number of syncs started
 *   loop
 *
 * In order to be scheduled by the scheduler, an entity must inherit from
 * either rt_sched_actor or rt_sched_sync.
 *
 * Currently, only actors get priority scheduled.  All synchronizers handled
 * by this scheduler run during every pass of the scheduler.  This was done
 * for a couple of reasons:  first, actors in an rtsync system will generally
 * outnumber synchronizers by a large degree so the performance hit is
 * minimal; second, the priority level of a synchronizer can only be properly
 * determined by running the synchronizer and therefore we cannot come up
 * with a proper implementation that does not involve constrantly running the
 * synchronizer threads to poll the messages contained within.
 */
void* scheduler(void*)
{
    using namespace std;

    // priority = time left to deadline
    // priority level = log(time left to deadline)/log(32)

    // Scheduler constants -- user configurable later on
    const int LEVELS = numeric_limits<unsigned>::digits;

    // time quantums for the actors and synchronizers (microseconds)
    const int ACTOR_QUANTUM = 400000;
    const int SYNC_QUANTUM = 320000;

    // populate the vector of sync threads
    vector<sched_entity_sync> syncs;
    for(std::map<pthread_t, rt_sched_sync*>::iterator it
            = cond_table_sync.begin();
        it != cond_table_sync.end(); ++it)
    {
        syncs.push_back(sched_entity_sync(&it->first, it->second));
    }

    // populate the multi-level structure which holds actor threads
    deque< deque<sched_entity_actor> > queue(LEVELS);
    for(std::map<pthread_t, rt_sched_actor*>::iterator it
            = cond_table_actor.begin();
        it != cond_table_actor.end(); ++it)
    {
        queue[0].push_back(sched_entity_actor(&it->first, it->second));
    }

    // bit of a kludge - indicates that we are on the first pass of the
    // scheduler
    bool first = true;

    // number of synchronizers handled by this scheduler will never change
    // so obtain a const_iterator to the end now
    const vector<sched_entity_sync>::const_iterator S_END
        = syncs.end();

    // start of main scheduler execution
    for(;;)
    {
        int outer_index=LEVELS-1;

        // Recalculate priorities
        rttime_t base = rt_gettime();
        for(deque< deque<sched_entity_actor> >::iterator outer = queue.begin();
            outer != queue.end(); ++outer)
        {
            unsigned n = 0, size = outer->size();
            for(deque<sched_entity_actor>::iterator it = outer->begin();
                n < size and it != outer->end(); ++n)
            {
                if (it->second->rt_sched_been_sched)
                    continue;
                uint32_t millis = 0;
                rttime_t deadline = it->second->getpriority();

                if(deadline >= base)
                {
                    deadline -= base;
                    millis = deadline.millis(); // ensure no overflow in rttime_t
                }

                it->second->rt_sched_been_sched = true;
                if(millis < 10)
                {
                    queue[0].push_back(*it);
                    outer_index = 0;
                }
                else
                {
                    int level = std::min((int) log2(millis), LEVELS-1);
                    outer_index = std::min(outer_index, level);
                    queue[level].push_back(*it);
                }

                it = outer->erase(it);
            }
        } // done recalculating priorities

        // if not on first pass of scheduler execution then synchronizer
        // threads have been running during recalculation of new actor
        // priority levels - so let's stop them now.
        if (!first)
        {
            for(vector<sched_entity_sync>::iterator cur = syncs.begin();
                cur != S_END; ++cur)
            {
                rt_sched_sync& sched_ref = *(cur->second);
                sched_ref.rt_sched_bool = false;
                pthread_kill(*(cur->first),SIGUSR2);
            }
        }
        else
            first = false;

        const deque<sched_entity_actor>::const_iterator END =
            queue[outer_index].end();

        // Iterate over highest-priority actor queue; wake up all the threads
        for(deque<sched_entity_actor>::iterator cur = queue[outer_index].begin();
            cur != END; ++cur)
        {
            rt_sched_actor& sched_ref = *(cur->second);
            sched_ref.rt_sched_bool = true;
            pthread_cond_signal(&sched_ref.rt_sched_token);
        }

        // wait
        usleep(ACTOR_QUANTUM * queue[outer_index].size());

        // Put actor threads back to sleep
        for(deque<sched_entity_actor>::iterator cur = queue[outer_index].begin();
            cur != END; ++cur)
        {
            rt_sched_actor& sched_ref = *(cur->second);
            sched_ref.rt_sched_bool = false;
            pthread_kill(*(cur->first),SIGUSR1);
        }

        // Run all of the synchronizers
        for(vector<sched_entity_sync>::iterator cur = syncs.begin();
            cur != S_END; ++cur)
        {
            rt_sched_sync& sched_ref = *(cur->second);
            sched_ref.rt_sched_bool = true;
            pthread_cond_signal(&sched_ref.rt_sched_token);
        }
        usleep(SYNC_QUANTUM * syncs.size());

        // Clear been_sched flag from the actor threads
        // (used during calc of priorities to indicate that we have already
        //  re-prioritized a thread)
        for(deque< deque<sched_entity_actor> >::iterator outer = queue.begin();
            outer != queue.end(); ++outer)
        {
            for(deque<sched_entity_actor>::iterator it = outer->begin();
                    it != outer->end(); ++it)
            {
                it->second->rt_sched_been_sched = false;
            }
        }
    }
}

// Function to start the scheduler
pthread_t sched_begin()
{
    pthread_t sched;
    pthread_create(&sched, 0, scheduler, 0);
    return sched;
}
