/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "synchronizer.h"
#include "postmaster.h"
#include "logger.h"

#include <string>
#include <algorithm>

#include "timing.h"

synchronizer::synchronizer(const syncref& ref, postmaster& p)
: name(ref), PostMaster(p)
{ }

// Receive a message, put it in the incoming queue
void synchronizer::receive_msg(message m)
{

    rt_log << name << " received message " << m.serial
           << " at " << rt_gettime() << "\n";
    rt_log << m << "\n";
    m.sync = name;
    m.timestamp.dispatch = rt_gettime();

    m_incoming.lock();
    m_incoming.push_back(m);
    m_incoming.unlock();

    rt_log << name << " stored message " << m.serial << ":\n"
           << m << "\n";
}

// Method for actors to request a message; the synchronizer shall send the
// highest priority message cached for that particular actor
message synchronizer::request_msg(const actorref& a)
{
    using namespace std;

    rt_log << a << " requesting message from " << name << "\n";

    auto_ptr<message> m_ptr;

    msg_cache.lock();
    // Check for cached message
    cached_msg& cm = msg_cache[a];
    pthread_mutex_t* mptr = msg_cache.get_pthread_mutex();
    // Wait until there is a message available
    while(!cm.avail)
        pthread_cond_wait(&cm.wait_var, mptr);
    message m = cm.msg;
    if (PostMaster.am_enabled(a, m.callee))
    {
        m_ptr.reset(new message(m));
        cm.avail = false;
    }
    // If message is disabled, return a null message
    else
        m_ptr.reset(new message(null_msg));
    msg_cache.unlock();

    rt_log << name << " relaying message from " << m_ptr->sender << " to "
           << m_ptr->receiver;
    if(*m_ptr == null_msg)
    {
        rt_log << " ... failed! " << m.callee << " disabled\n";
        return null_msg;
    }
    else
        rt_log << " ... OK\n";

    return *m_ptr;
}

// Method for actors to acknowledge that processing of message is complete
void synchronizer::ack_msg(message m)
{
    actorref& a = m.receiver;
    rt_log << a << " notified " << name << " of processing completion\n";
}

// Polling functionality below

// To pass to pthread_create
void* start_polling_func_s(synchronizer* s)
{
    s->poll_master();
    return 0;
}

// Start the polling thread; each derived constructor shall call this once
// its members are initialized
void synchronizer::start_polling()
{
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

    int result = pthread_create(&m_pollthread, &attr,
                                (void* (*)(void*)) start_polling_func_s,
                                (void*) this);

#ifdef USE_SCHED
    sched_register(m_pollthread);
#endif

    rt_log << name << " is " << m_pollthread << "\n";

    // Very bad if below happens
    if(result)
        throw std::string("Error: Synchronizer failed to create polling thread");
}

// Kill the polling thread; each derived destructor shall call this before
// destroying any of its members
void synchronizer::end_polling()
{
    pthread_kill(m_pollthread, 9);
}

// Master polling function; iterates over internal data structures.
void synchronizer::poll_master()
{
    // Outgoing message queue
    typedef std::map<actorref, std::vector<message> > msg_map;
    msg_map m_outgoing;
    std::list<message> local_incoming;

    rt_log << name << " starts its polling ...\n";

    while(true)
    {
        try
        {
            // Dump exposed incoming queue into the local one
            m_incoming.lock();
            local_incoming.insert(local_incoming.end(),
                                  m_incoming.begin(), m_incoming.end());
            m_incoming.clear();
            m_incoming.unlock();

            std::list<message>::iterator list_it = local_incoming.begin();

            // First loop checks the synchronizer's incoming queue to see
            // if there are any messages that (after checking constraints) can
            // be moved to the synchonizer's outgoing queue.
            while(list_it != local_incoming.end())
            {
                int res;
                try {
                    res = poll_msg(*list_it);
                }
                catch(...) {
                    list_it = local_incoming.erase(list_it);
                    throw;
                }
                if(res == REMOVE)
                {
                    rt_log << name << " removing message " << list_it->serial
                           << "\n";
                    m_outgoing[list_it->receiver].push_back(*list_it);
                    PostMaster.am_register_msg(list_it->receiver, name,
                                               list_it->must_run_before);
                    list_it = local_incoming.erase(list_it);
                }
                else
                {
                    ++list_it;
                }
            }

            poll_triggers();

            // This loop checks the messages currently waiting in the outgoing
            // queue against the constraints to see if we have a timing
            // violation.

            for(msg_map::iterator map_it = m_outgoing.begin();
                map_it != m_outgoing.end(); ++map_it)
            {
                for(std::vector<message>::iterator it = map_it->second.begin();
                    it != map_it->second.end();)
                {
                    try {
                        poll_msg(*it);
                    }
                    catch(...) {
                        std::swap(*it, map_it->second.back());
                        map_it->second.pop_back();
                        throw;
                    }

                    msg_cache.lock();
                    cached_msg& cm = msg_cache[it->receiver];

                    if (!cm.avail)
                    {
                        cm.msg = *it;
                        cm.avail = true;
                        msg_cache.unlock();
                        pthread_cond_signal(&cm.wait_var);
                        std::swap(*it, map_it->second.back());
                        map_it->second.pop_back();
                    }
                    else
                    {
                        if (*it < cm.msg)
                        {
                            std::swap(*it, cm.msg);
                        }
                        msg_cache.unlock();
                        ++it;
                    }
                }
            } // end map iteration

            msg_cache.lock();
            for(lock_map<actorref,cached_msg>::iterator it = msg_cache.begin();
                it != msg_cache.end(); ++it)
            {
                if (it->second.avail)
                {
                    try {
                        poll_msg(it->second.msg);
                    }
                    catch(...) {
                        it->second.avail = false;
                        msg_cache.unlock();
                        throw;
                    }
                }
            }
            msg_cache.unlock();
        }
        catch(const timing_exception& te)
        {
            rt_log << "***Timing Exception: \n" << te << "\n\n";
            int val = exception_handler(te);
            if(val == TERMINATE)
                PostMaster.terminate_system();
            else if (val == KILLNODE)
                PostMaster.terminate_node();
        }
        // Terrible, terrible, ugly, but sadly vital hack. The thread must
        // sleep for a while to ensure that, in the event of the majority of
        // the queues being empty, there is a significant chance that, when
        // a signal arrives from the scheduler, the thread does not presently
        // hold a mutex.
        usleep(10000);
    }
} // end poll_master
