/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef SYNCHRONIZER_H
#define SYNCHRONIZER_H


#include <list>
#include <map>
#include <memory>
#include <vector>

#include <pthread.h>

#include "messages.h"
#include "actionlistener.h"
#include "schedulable.h"
#include "reftypes.h"
#include "lock_map.h"
#include "lock_list.h"

class postmaster;


// The synchronizer is the abstract base class for synchronizers generated
// from user code (see docs/sync_concepts for an overview). Generated
// synchronizers shall provide suitable constructors and implement three pure
// virtual functions: poll_triggers(), poll_msg(), and exception_handler().
//    A synchronizer functions by associating itself with a postmaster,
// setting up a polling loop to iterate over and process its internal data
// structures, and registering itself as a listener and synchronizer to
// various actors.
class synchronizer : public actionlistener, public rt_sched_sync
{
    // Simple structure to handle cached messages
    struct cached_msg
    {
        bool avail; // Is there an available message?
        pthread_cond_t wait_var;
        message msg;

        cached_msg() : avail(false) {
            pthread_cond_init(&wait_var, 0);
        }

        ~cached_msg() {
            pthread_cond_destroy(&wait_var);
        }
    };

    // Thread id of polling thread
    pthread_t m_pollthread;

    // In order to avoid superfluous mutex locks on the master internal
    // structures, an associate map of the highest priority message for each
    // recipient actor is stored here.
    //   Note: Once a cached_msg has been inserted into this structure, it
    // shall _not_ be erased even after it has been processed, but rather it
    // shall be flagged as unavailable.
    lock_map<actorref, cached_msg> msg_cache;

    // Incoming message queue
    lock_list<message> m_incoming;

    // Master polling function; iterates over internal data structures.
    virtual void poll_master();

    // Utility function: Wraps member function to start with pthread_create
    friend void* start_polling_func_s(synchronizer* s);

protected:
    syncref name;
    postmaster& PostMaster;

    // Start the polling thread; each derived constructor shall call this once
    // its members are initialized
    virtual void start_polling();
    // Kill the polling thread; each derived destructor shall call this before
    // destroying any of its members
    virtual void end_polling();


    // The functions below should not be called by users or derived classes;
    // the thread creation shall take care of this.

    // Poll enable/disable triggers; do not call this explicitly. This
    // function shall examine certain conditions and set and propagate the
    // enabled/disabled status of actor methods appropriately.
    //   This shall poll each trigger only once; it will be called in a loop.
    virtual void poll_triggers() = 0;

    // Poll a message; do not call. This shall examine a message, set its
    // deadlines appropriately, and throw exceptions if timing constraints
    // are violated. The exception behaviour
    //   Poll each message once; this will be called in a loop.
    //   The result shall be either DEFAULT or REMOVE, the latter meaning
    // that the message is ready to run and can be dispatched.
    enum poll_result { DEFAULT, REMOVE };
    virtual poll_result poll_msg(message& m) = 0;

    // Exception handler used to process timing exceptions thrown by
    // poll_msg(). This can execute arbitrary code, but shall return one of
    // the symbolic constants below.
    virtual int exception_handler(const timing_exception&) = 0;

    // Symbolic constants for exception responses
    static const int TERMINATE = 99;     // Terminate the system
    static const int KILLNODE  = 443556; // Terminate the local node
    static const int CONTINUE  = 0;      // Proceed as before

public:
    synchronizer(const syncref& ref, postmaster& p);
    virtual ~synchronizer() { }

    // Receive a message, put it in the incoming queue and set the received
    // timestamp
    virtual void receive_msg(message m);

    // Method for actors to request a message; the synchronizer shall send the
    // highest priority message cached for that particular actor
    virtual message request_msg(const actorref& a);

    // Method for actors to acknowledge that processing of message is complete
    virtual void ack_msg(message m);

};

#endif // #define SYNCHRONIZER_H
