/*
 * RTSync: real-time compiler and runtime environment
 * Copyright (C) 2004  Stefan D. Bruda, Petter Haggholm, and Scott Stoddard
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named LICENSE; if not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef TIMING_H
#define TIMING_H


#include <sys/time.h>
#include <boost/cstdint.hpp>
#include <boost/integer_traits.hpp>


// This header shall define a suitable data type for storing time as rttime_t,
// a function rt_gettime() to retrieve the current time, and operators for
// comparing instances of this type by arithmetic comparisons:
//   > < == >= <= !=
// For debugging purposes, a stream insertion operator shall also be provided.
// In case of non-POD structures, suitable constructors/destructors/assignment
// operations shall be provided.
// A constant symbol rt_inf conceptually representing infinite time is
// declared.

struct rttime_t
{
    time_t tv_sec;
    suseconds_t tv_usec;

    rttime_t() : tv_sec(0), tv_usec(0) { }
    rttime_t(const rttime_t& t) : tv_sec(t.tv_sec), tv_usec(t.tv_usec) { }
    rttime_t(const timeval& t) : tv_sec(t.tv_sec), tv_usec(t.tv_usec) { }
    rttime_t(long s, long ms) : tv_sec(s), tv_usec(ms*1000) { }

    const rttime_t& operator=(const rttime_t& t)
    {
        tv_sec = t.tv_sec;
        tv_usec = t.tv_usec;
        return *this;
    }

    const rttime_t& operator+=(const rttime_t& t)
    {
        tv_sec  += t.tv_sec;
        tv_usec += t.tv_usec;

        int tmp = t.tv_usec/1000000;
        tv_sec  += tmp;
        tv_usec -= tmp * 1000000;

        return *this;
    }

    const rttime_t& operator-=(const rttime_t& t)
    {
        tv_sec  -= t.tv_sec;
        tv_usec -= t.tv_usec;

        if(tv_usec < 0)
        {
            tv_usec += 1000000;
            tv_sec -= 1;
        }

        return *this;
    }

    // Returns a 32-bit number representing the millisecond interpretation or,
    // if the time is too great, std::numeric_limits<uint32_t>::max()
    boost::uint32_t millis() const
    {
        unsigned long t = (unsigned long) tv_sec;
        if(boost::integer_traits<unsigned long>::const_max - tv_usec/1000000
           < t)
        {
            t = boost::integer_traits<unsigned long>::const_max;
        }
        else
            t += tv_usec/1000000;

        if(t >= boost::integer_traits<uint32_t>::const_max/1000)
            return boost::integer_traits<uint32_t>::const_max;
        else
            return (uint32_t) (t*1000);
    }
};


std::istream& operator>>(std::istream& is, rttime_t& t);
std::ostream& operator<<(std::ostream& os, const rttime_t& t);

extern const rttime_t rt_inf;     // infinity
extern const rttime_t rt_zero;    // 0 as an rttime_t
extern const rttime_t rt_startup; // node startup


inline rttime_t rt_gettime()
{
    timeval t;
    gettimeofday(&t, 0);
    return t;
}


// Arithmetic

inline rttime_t operator+(const rttime_t& t1, const rttime_t& t2)
{
    rttime_t r(t1);
    r += t2;
    return r;
}


inline rttime_t operator-(const rttime_t& t1, const rttime_t& t2)
{
    rttime_t r(t1);
    r -= t2;
    return r;
}


// Comparison operators

inline bool operator<(const rttime_t& t1, const rttime_t& t2)
{
    return (t1.tv_sec < t2.tv_sec) or
           (t1.tv_sec == t2.tv_sec and t1.tv_usec < t2.tv_usec);
}

inline bool operator>(const rttime_t& t1, const rttime_t& t2)
{
    return (t1.tv_sec > t2.tv_sec) or
           (t1.tv_sec == t2.tv_sec and t1.tv_usec > t2.tv_usec);
}

inline bool operator==(const rttime_t& t1, const rttime_t& t2)
{
    return (t1.tv_sec == t2.tv_sec and t1.tv_usec == t2.tv_usec);
}

inline bool operator<=(const rttime_t& t1, const rttime_t& t2)
{
    return (t1.tv_sec < t2.tv_sec) or
           (t1.tv_sec == t2.tv_sec and t1.tv_usec <= t2.tv_usec);
}

inline bool operator>=(const rttime_t& t1, const rttime_t& t2)
{
    return (t1.tv_sec > t2.tv_sec) or
           (t1.tv_sec == t2.tv_sec and t1.tv_usec >= t2.tv_usec);
}

inline bool operator!=(const rttime_t& t1, const rttime_t& t2)
{
    return (t1.tv_sec != t2.tv_sec or t1.tv_usec != t2.tv_usec);
}

#endif // #define TIMING_H
